from __future__ import annotations

import random
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Protocol

from ..paths import default_runtime_dir

if TYPE_CHECKING:
    from grim.assets import LogoAssets, PaqTextureCache
    from grim.audio import AudioState
    from grim.config import CrimsonConfig
    from grim.console import ConsoleState
    from grim.geom import Vec2
    from grim.terrain_render import GroundRenderer

    from ..modes.quest_mode import QuestRunOutcome
    from ..net.net_runtime import NetRuntime
    from ..net.runtime import LanRuntime
    from ..persistence.save_status import GameStatus


@dataclass(frozen=True, slots=True)
class GameConfig:
    base_dir: Path = field(default_factory=default_runtime_dir)
    assets_dir: Path | None = None
    width: int | None = None
    height: int | None = None
    fps: int = 60
    seed: int | None = None
    demo_enabled: bool = False
    no_intro: bool = False
    debug: bool = False
    preserve_bugs: bool = False
    pending_net_session: "PendingNetSession | None" = None
    pending_lan_session: "PendingLanSession | None" = None


LanSessionMode = Literal["survival", "rush", "quests"]
LanSessionRole = Literal["host", "join"]
NetSessionMode = LanSessionMode
NetSessionRole = LanSessionRole
NetcodeMode = Literal["rollback", "lockstep_legacy"]


@dataclass(frozen=True, slots=True)
class LanSessionConfig:
    mode: LanSessionMode
    player_count: int = 1
    quest_level: str = ""
    bind_host: str = "0.0.0.0"
    relay_host: str = ""
    relay_port: int = 31993
    room_code: str = ""
    host_ip: str = ""
    port: int = 31993
    netcode_mode: NetcodeMode = "rollback"
    rollback_max_ticks: int = 8
    reconnect_timeout_ms: int = 15_000
    input_delay_ticks: int = 1
    preserve_bugs: bool = False

    def resolved_relay_host(self) -> str:
        host = str(self.relay_host).strip()
        if host:
            return host
        fallback = str(self.host_ip).strip()
        if fallback:
            return fallback
        return "127.0.0.1"

    def resolved_relay_port(self) -> int:
        relay_port = int(self.relay_port)
        if relay_port > 0:
            return relay_port
        return int(self.port)


@dataclass(slots=True)
class PendingLanSession:
    role: LanSessionRole
    config: LanSessionConfig
    auto_start: bool = False
    started: bool = False
    error: str = ""


NetSessionConfig = LanSessionConfig
PendingNetSession = PendingLanSession


@dataclass(slots=True)
class HighScoresRequest:
    game_mode_id: int
    quest_stage_major: int = 0
    quest_stage_minor: int = 0
    highlight_rank: int | None = None


class FrontView(Protocol):
    def open(self) -> None: ...

    def close(self) -> None: ...

    def update(self, dt: float) -> None: ...

    def draw(self) -> None: ...

    def take_action(self) -> str | None: ...


class PauseBackground(Protocol):
    def draw_pause_background(self, *, entity_alpha: float = 1.0) -> None: ...


@dataclass(slots=True)
class GameState:
    base_dir: Path
    assets_dir: Path
    rng: random.Random
    config: CrimsonConfig
    status: GameStatus
    console: ConsoleState
    demo_enabled: bool
    preserve_bugs: bool
    logos: LogoAssets | None
    texture_cache: PaqTextureCache | None
    audio: AudioState | None
    resource_paq: Path
    session_start: float
    skip_intro: bool = False
    gamma_ramp: float = 1.0
    snd_freq_adjustment_enabled: bool = True
    menu_ground: GroundRenderer | None = None
    menu_ground_camera: Vec2 | None = None
    menu_sign_locked: bool = False
    stats_menu_easter_egg_roll: int = -1
    pause_background: PauseBackground | None = None
    pending_net_session: PendingNetSession | None = None
    net_runtime: "NetRuntime | LanRuntime | None" = None
    net_in_lobby: bool = False
    net_waiting_for_players: bool = False
    net_expected_players: int = 1
    net_connected_players: int = 1
    net_desync_count: int = 0
    net_resync_failure_count: int = 0
    net_last_error: str = ""
    pending_lan_session: PendingLanSession | None = None
    lan_runtime: "NetRuntime | LanRuntime | None" = None
    lan_in_lobby: bool = False
    lan_waiting_for_players: bool = False
    lan_expected_players: int = 1
    lan_connected_players: int = 1
    lan_desync_count: int = 0
    lan_resync_failure_count: int = 0
    lan_last_error: str = ""
    pending_quest_level: str | None = None
    pending_high_scores: HighScoresRequest | None = None
    quest_outcome: QuestRunOutcome | None = None
    quest_fail_retry_count: int = 0
    terrain_regenerate_requested: bool = False
    survival_elapsed_ms: float = 0.0
    demo_trial_elapsed_ms: int = 0
    quit_requested: bool = False
    screen_fade_alpha: float = 0.0
    screen_fade_ramp: bool = False


__all__ = [
    "FrontView",
    "GameConfig",
    "GameState",
    "HighScoresRequest",
    "LanSessionConfig",
    "NetSessionConfig",
    "NetcodeMode",
    "PendingNetSession",
    "PendingLanSession",
    "PauseBackground",
]



from __future__ import annotations

from .bootstrap import AppliedReplayBootstrap, ReplayBootstrapError, apply_replay_bootstrap
from .codec import ReplayCodecError, dump_replay, dump_replay_file, load_replay, load_replay_file
from .input_codec import pack_player_input, pack_tick_inputs, unpack_player_input, unpack_tick_inputs
from .recorder import ReplayRecorder
from .types import (
    FIRE_DOWN_FLAG,
    FIRE_PRESSED_FLAG,
    MOVE_BACKWARD_FLAG,
    MOVE_FORWARD_FLAG,
    MOVE_KEYS_PRESENT_FLAG,
    RELOAD_PRESSED_FLAG,
    TURN_LEFT_FLAG,
    TURN_RIGHT_FLAG,
    PackedPlayerInput,
    PackedTickInputs,
    PerkMenuOpenEvent,
    PerkPickEvent,
    Replay,
    ReplayHeader,
    ReplayStatusSnapshot,
    UnknownEvent,
    pack_input_flags,
    unpack_input_flags,
    unpack_input_move_key_flags,
    unpack_packed_player_input,
)
from .versioning import ReplayGameVersionWarning, warn_on_game_version_mismatch

__all__ = [
    "FIRE_DOWN_FLAG",
    "FIRE_PRESSED_FLAG",
    "MOVE_BACKWARD_FLAG",
    "MOVE_FORWARD_FLAG",
    "MOVE_KEYS_PRESENT_FLAG",
    "RELOAD_PRESSED_FLAG",
    "TURN_LEFT_FLAG",
    "TURN_RIGHT_FLAG",
    "PackedPlayerInput",
    "PackedTickInputs",
    "PerkMenuOpenEvent",
    "PerkPickEvent",
    "Replay",
    "ReplayCodecError",
    "AppliedReplayBootstrap",
    "ReplayBootstrapError",
    "ReplayHeader",
    "ReplayRecorder",
    "ReplayStatusSnapshot",
    "UnknownEvent",
    "ReplayGameVersionWarning",
    "dump_replay",
    "dump_replay_file",
    "load_replay",
    "load_replay_file",
    "pack_player_input",
    "apply_replay_bootstrap",
    "pack_input_flags",
    "pack_tick_inputs",
    "unpack_player_input",
    "unpack_packed_player_input",
    "unpack_input_flags",
    "unpack_input_move_key_flags",
    "unpack_tick_inputs",
    "warn_on_game_version_mismatch",
]



"""Headless oracle mode for differential testing.

Runs the game simulation without rendering, accepts inputs from a JSON file,
and emits game state to stdout each frame for comparison with other implementations.
"""

from __future__ import annotations

import hashlib
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from grim.geom import Vec2

from .sim.driver.setup import build_damage_scale_by_type
from .sim.input import PlayerInput
from .sim.sessions import SurvivalDeterministicSession
from .sim.state_types import PlayerState
from .sim.world_state import WorldState


class OutputMode:
    """Output modes for oracle state emission."""

    FULL = "full"  # Complete state every sample
    SUMMARY = "summary"  # Score, kills, player pos/health only
    HASH = "hash"  # SHA256 hash of full state for fast comparison
    CHECKPOINTS = "checkpoints"  # Only on significant events


@dataclass(frozen=True, slots=True)
class OracleConfig:
    """Configuration for headless oracle mode."""

    seed: int
    input_file: Path | None
    max_frames: int = 36000  # 10 minutes at 60fps
    frame_rate: int = 60
    sample_rate: int = 1  # Emit state every N frames (1 = every frame, 60 = once per second)
    output_mode: str = OutputMode.SUMMARY
    preserve_bugs: bool = False


@dataclass(slots=True)
class FrameInput:
    """Input for a single frame."""

    frame: int
    move: Vec2 = field(default_factory=Vec2)
    aim: Vec2 = field(default_factory=Vec2)
    fire_down: bool = False
    fire_pressed: bool = False
    reload_pressed: bool = False


def load_inputs(path: Path) -> list[FrameInput]:
    """Load input sequence from JSON file.

    Expected format:
    {
        "frames": [
            {"frame": 0, "move_x": 1.0, "move_y": 0.0, "aim": [100, 200], "fire_down": true},
            {"frame": 60, "move_x": 0.0, "move_y": -1.0, "fire_pressed": true},
            ...
        ]
    }
    """
    data = json.loads(path.read_text())
    inputs: list[FrameInput] = []
    for entry in data.get("frames", []):
        raw_aim = entry.get("aim", [0.0, 0.0])
        if not isinstance(raw_aim, list) or len(raw_aim) != 2:
            raise ValueError(f"frame {entry.get('frame', 0)} has invalid aim payload: expected [x, y]")
        inputs.append(
            FrameInput(
                frame=int(entry.get("frame", 0)),
                move=Vec2(float(entry.get("move_x", 0.0)), float(entry.get("move_y", 0.0))),
                aim=Vec2(float(raw_aim[0]), float(raw_aim[1])),
                fire_down=bool(entry.get("fire_down", False)),
                fire_pressed=bool(entry.get("fire_pressed", False)),
                reload_pressed=bool(entry.get("reload_pressed", False)),
            ),
        )
    return sorted(inputs, key=lambda i: i.frame)


def export_player_state(player: PlayerState) -> dict[str, Any]:
    """Export player state to JSON-serializable dict."""
    return {
        "index": int(player.index),
        "pos": player.pos.to_dict(ndigits=4),
        "health": round(float(player.health), 4),
        "weapon_id": int(player.weapon_id),
        "ammo": round(float(player.ammo), 4),
        "experience": int(player.experience),
        "level": int(player.level),
        "reload_active": bool(player.reload_active),
        "heading": round(float(player.heading), 4),
        "aim_heading": round(float(player.aim_heading), 4),
    }


def export_creature_state(creature: Any) -> dict[str, Any]:
    """Export creature state to JSON-serializable dict."""
    return {
        "id": int(creature.id) if hasattr(creature, "id") else -1,
        "type_id": int(creature.type_id),
        **creature.pos.to_dict(ndigits=4),
        "hp": round(float(creature.hp), 4),
        "active": bool(creature.active),
    }


def export_bonus_state(bonus: Any) -> dict[str, Any]:
    """Export bonus state to JSON-serializable dict."""
    return {
        "bonus_id": int(bonus.bonus_id),
        "pos": bonus.pos.to_dict(ndigits=4),
        "time_left": round(float(bonus.time_left), 4),
        "picked": bool(bonus.picked),
    }


def export_projectile_state(proj: Any) -> dict[str, Any]:
    """Export projectile state to JSON-serializable dict."""
    return {
        "type_id": int(proj.type_id),
        "x": round(proj.x, 4),
        "y": round(proj.y, 4),
        "active": bool(proj.active),
    }


def export_game_state_full(
    frame: int,
    world_state: WorldState,
    players: list[PlayerState],
    rng_state: int,
    elapsed_ms: float,
    command_hash: str,
) -> dict[str, Any]:
    """Export complete game state for a frame."""
    state = world_state.state

    # Collect active creatures
    creatures = []
    for creature in world_state.creatures.entries:
        if creature.active:
            creatures.append(export_creature_state(creature))

    # Collect active bonuses
    bonuses = []
    for bonus in state.bonus_pool.iter_active():
        bonuses.append(export_bonus_state(bonus))

    # Collect active projectiles
    projectiles = []
    for proj in state.projectiles.entries:
        if proj.active:
            projectiles.append(export_projectile_state(proj))

    # Score is player experience, kills tracked on creatures pool
    total_experience = sum(p.experience for p in players)
    kill_count = world_state.creatures.kill_count

    return {
        "frame": frame,
        "rng_state": rng_state,
        "command_hash": str(command_hash),
        "elapsed_ms": round(elapsed_ms, 4),
        "score": int(total_experience),
        "kills": int(kill_count),
        "players": [export_player_state(p) for p in players],
        "creatures": creatures,
        "bonuses": bonuses,
        "projectiles": projectiles,
        "bonus_timers": {
            "weapon_power_up": round(float(state.bonuses.weapon_power_up), 4),
            "reflex_boost": round(float(state.bonuses.reflex_boost), 4),
            "freeze": round(float(state.bonuses.freeze), 4),
        },
    }


def export_game_state_summary(
    frame: int,
    world_state: WorldState,
    players: list[PlayerState],
    rng_state: int,
    elapsed_ms: float,
    command_hash: str,
) -> dict[str, Any]:
    """Export minimal game state for fast comparison."""
    total_experience = sum(p.experience for p in players)
    kill_count = world_state.creatures.kill_count
    creature_count = sum(1 for c in world_state.creatures.entries if c.active)

    return {
        "frame": frame,
        "rng_state": rng_state,
        "command_hash": str(command_hash),
        "elapsed_ms": round(elapsed_ms, 4),
        "score": int(total_experience),
        "kills": int(kill_count),
        "creature_count": creature_count,
        "players": [
            {
                "pos": p.pos.to_dict(ndigits=2),
                "health": round(float(p.health), 2),
                "weapon_id": int(p.weapon_id),
                "level": int(p.level),
            }
            for p in players
        ],
    }


def export_game_state_hash(
    frame: int,
    world_state: WorldState,
    players: list[PlayerState],
    rng_state: int,
    elapsed_ms: float,
    command_hash: str,
) -> dict[str, Any]:
    """Export hash of game state for ultra-fast comparison."""
    # Get full state and hash it
    full_state = export_game_state_full(frame, world_state, players, rng_state, elapsed_ms, command_hash)
    # Remove frame from hash computation (it's metadata)
    hashable = {k: v for k, v in full_state.items() if k != "frame"}
    state_bytes = json.dumps(hashable, sort_keys=True).encode()
    state_hash = hashlib.sha256(state_bytes).hexdigest()[:16]

    return {
        "frame": frame,
        "hash": state_hash,
        "command_hash": str(command_hash),
        "score": full_state["score"],
        "kills": full_state["kills"],
    }


@dataclass(slots=True)
class CheckpointTracker:
    """Track significant events for checkpoint-only output."""

    last_score: int = 0
    last_kills: int = 0
    last_level: int = 1
    last_health: float = 100.0
    last_weapon_id: int = 1

    def check_and_update(self, players: list[PlayerState], world_state: WorldState) -> bool:
        """Return True if any significant change occurred."""
        score = sum(p.experience for p in players)
        kills = world_state.creatures.kill_count
        level = players[0].level if players else 1
        health = players[0].health if players else 0.0
        weapon_id = players[0].weapon_id if players else 1

        changed = (
            score != self.last_score
            or kills != self.last_kills
            or level != self.last_level
            or int(health) != int(self.last_health)  # Only trigger on integer health change
            or weapon_id != self.last_weapon_id
        )

        if changed:
            self.last_score = score
            self.last_kills = kills
            self.last_level = level
            self.last_health = health
            self.last_weapon_id = weapon_id

        return changed


def run_headless(config: OracleConfig) -> None:
    """Run the game in headless mode, emitting state JSON each frame."""
    from .effects import FxQueue, FxQueueRotated

    # Build world state
    world_state = WorldState.build(
        world_size=1024.0,
        demo_mode_active=False,
        hardcore=False,
        difficulty_level=0,
        preserve_bugs=bool(config.preserve_bugs),
    )

    # Initialize with seed
    world_state.state.rng.srand(config.seed)

    # Set up player at center
    players = world_state.players
    if not players:
        from .sim.state_types import PlayerState
        from .weapon_runtime import weapon_assign_player

        player = PlayerState(index=0, pos=Vec2(512.0, 512.0))
        weapon_assign_player(player, 1)
        players.append(player)

    # Load inputs if provided
    inputs_by_frame: dict[int, FrameInput] = {}
    if config.input_file is not None:
        for inp in load_inputs(config.input_file):
            inputs_by_frame[inp.frame] = inp

    dt = 1.0 / float(config.frame_rate)
    current_input = FrameInput(frame=0)

    # Headless deterministic step still routes through sim/presentation queues.
    fx_queue = FxQueue()
    fx_queue_rotated = FxQueueRotated()
    session = SurvivalDeterministicSession(
        world=world_state,
        world_size=1024.0,
        damage_scale_by_type=build_damage_scale_by_type(),
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        detail_preset=5,
        fx_toggle=0,
        game_tune_started=False,
        auto_pick_perks=True,
        demo_mode_active=False,
        perk_progression_enabled=True,
        clear_fx_queues_each_tick=True,
    )
    last_command_hash = ""

    # Checkpoint tracker for event-driven output
    checkpoint_tracker = CheckpointTracker()

    # Select export function based on output mode
    export_fn = {
        OutputMode.FULL: export_game_state_full,
        OutputMode.SUMMARY: export_game_state_summary,
        OutputMode.HASH: export_game_state_hash,
        OutputMode.CHECKPOINTS: export_game_state_summary,  # Same format, different trigger
    }.get(config.output_mode, export_game_state_summary)

    for frame in range(config.max_frames):
        # Update current input if we have one for this frame
        if frame in inputs_by_frame:
            current_input = inputs_by_frame[frame]

        # Convert to PlayerInput
        player_inputs = [
            PlayerInput(
                move=current_input.move,
                aim=current_input.aim,
                fire_down=current_input.fire_down,
                fire_pressed=current_input.fire_pressed,
                reload_pressed=current_input.reload_pressed,
            ),
        ]

        tick = session.step_tick(dt_frame=dt, inputs=player_inputs)
        last_command_hash = str(tick.step.command_hash)
        elapsed_ms = float(session.elapsed_ms)

        # Determine if we should emit state this frame
        should_emit = False
        if config.output_mode == OutputMode.CHECKPOINTS:
            # Only emit on significant changes
            should_emit = checkpoint_tracker.check_and_update(players, world_state)
            # Always emit first and last frame
            if frame == 0:
                should_emit = True
        else:
            # Sample rate based emission
            should_emit = frame % config.sample_rate == 0

        if should_emit:
            state_json = export_fn(
                frame=frame,
                world_state=world_state,
                players=players,
                rng_state=world_state.state.rng.state,
                elapsed_ms=elapsed_ms,
                command_hash=str(last_command_hash),
            )
            print(json.dumps(state_json), flush=True)

        # Check if all players dead
        if all(p.health <= 0 for p in players):
            # Always emit final state on death
            if not should_emit:
                state_json = export_fn(
                    frame=frame,
                    world_state=world_state,
                    players=players,
                    rng_state=world_state.state.rng.state,
                    elapsed_ms=elapsed_ms,
                    command_hash=str(last_command_hash),
                )
                print(json.dumps(state_json), flush=True)
            break



from __future__ import annotations

import math
from collections.abc import Sequence

from grim.geom import Vec2

from ..types import _CREATURE_HITBOX_ALIVE, CreatureDamageApplier, Damageable, _SizeLike

_NATIVE_FIND_RADIUS_MARGIN_EPS = 0.001


def _hit_radius_for(creature: _SizeLike) -> float:
    """Approximate `creature_find_in_radius`/`creatures_apply_radius_damage` sizing.

    The native code compares `distance - radius < creature.size * 0.14285715 + 3.0`.
    """

    size = float(creature.size)
    return max(0.0, size * 0.14285715 + 3.0)


def _within_native_find_radius(*, origin: Vec2, target: Vec2, radius: float, target_size: float) -> bool:
    """Mirror native `creature_find_in_radius` / `player_find_in_radius` predicate.

    Native uses:
      sqrt(dx*dx + dy*dy) - radius < size * 0.14285715 + 3.0
    """

    dx = float(target.x) - float(origin.x)
    dy = float(target.y) - float(origin.y)
    radius_f = float(radius)
    size_margin = float(target_size) * 0.14285715 + 3.0
    max_axis_delta = float(radius_f) + float(size_margin) + _NATIVE_FIND_RADIUS_MARGIN_EPS
    # Fast reject for the common case where either axis already exceeds the
    # maximal accepted Euclidean radius.
    if abs(float(dx)) > float(max_axis_delta) or abs(float(dy)) > float(max_axis_delta):
        return False
    margin = math.sqrt(dx * dx + dy * dy) - float(radius_f) - float(size_margin)
    # Native uses x87-heavy float math in this path; a tiny epsilon avoids
    # branch flips on sub-millimetric drift from float32 replay state.
    return float(margin) < _NATIVE_FIND_RADIUS_MARGIN_EPS


def _creature_find_nearest_for_secondary(*, creatures: Sequence[Damageable], origin: Vec2) -> int:
    """Port of `creature_find_nearest(origin, -1, 0.0)` for homing secondary targets."""

    best_idx = 0
    best_dist_sq = 1_000_000.0
    max_index = min(len(creatures), 0x180)
    for idx in range(max_index):
        creature = creatures[idx]
        if not creature.active:
            continue
        if float(creature.hitbox_size) != _CREATURE_HITBOX_ALIVE:
            continue
        dist_sq = Vec2.distance_sq(origin, creature.pos)
        if dist_sq < best_dist_sq:
            best_dist_sq = dist_sq
            best_idx = idx
    return best_idx


def _apply_damage_to_creature(
    creatures: Sequence[Damageable],
    creature_index: int,
    damage: float,
    *,
    damage_type: int,
    impulse: Vec2,
    owner_id: int,
    apply_creature_damage: CreatureDamageApplier | None = None,
) -> None:
    if damage <= 0.0:
        return
    idx = int(creature_index)
    if not (0 <= idx < len(creatures)):
        return
    if apply_creature_damage is not None:
        apply_creature_damage(
            idx,
            float(damage),
            int(damage_type),
            impulse,
            int(owner_id),
        )
    else:
        creatures[idx].hp -= float(damage)


__all__ = [
    "_apply_damage_to_creature",
    "_creature_find_nearest_for_secondary",
    "_hit_radius_for",
    "_within_native_find_radius",
]



from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Protocol

import pyray as rl

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp
from grim.terrain_render import _maybe_alpha_test

from ...creatures.spawn import CreatureFlags, CreatureTypeId
from ...effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, SIZE_CODE_GRID, EffectId
from ...perks import PerkId
from ...perks.helpers import perk_active
from ...sim.world_defs import CREATURE_ANIM, CREATURE_ASSET
from ...ui.cursor import draw_aim_cursor
from ..frame import RenderFrame
from .constants import _RAD_TO_DEG, monster_vision_fade_alpha
from .mixin_base import WorldRendererMixinBase

if TYPE_CHECKING:
    from ...sim.state_types import PlayerState


class _CreatureDrawContract(Protocol):
    active: bool
    pos: Vec2
    size: float
    hitbox_size: float
    type_id: int
    flags: int
    plague_infected: bool


@dataclass(frozen=True, slots=True)
class WorldDrawContext:
    camera: Vec2 = field(default_factory=Vec2)
    view_scale: Vec2 = field(default_factory=lambda: Vec2(1.0, 1.0))
    scale: float = 1.0
    entity_alpha: float = 1.0
    trooper_texture: rl.Texture | None = None
    particles_texture: rl.Texture | None = None
    monster_vision: bool = False
    monster_vision_src: rl.Rectangle | None = None
    poison_src: rl.Rectangle | None = None


class WorldRendererDrawMixin(WorldRendererMixinBase):
    def draw(
        self,
        *,
        render_frame: RenderFrame | None = None,
        draw_aim_indicators: bool = True,
        entity_alpha: float = 1.0,
    ) -> None:
        self._render_frame = render_frame if render_frame is not None else self._world.build_render_frame()
        try:
            self._draw_with_active_frame(draw_aim_indicators=draw_aim_indicators, entity_alpha=entity_alpha)
        finally:
            self._render_frame = None

    def _draw_with_active_frame(self, *, draw_aim_indicators: bool = True, entity_alpha: float = 1.0) -> None:
        entity_alpha = clamp(float(entity_alpha), 0.0, 1.0)
        camera, view_scale, scale, screen_size, out_size = self._compute_view_transform()
        self._draw_background(camera=camera, screen_size=screen_size, out_size=out_size, view_scale=view_scale)
        if entity_alpha <= 1e-3:
            return

        with _maybe_alpha_test(bool(self._alpha_test_enabled())):
            ctx = self._build_draw_context(
                camera=camera,
                view_scale=view_scale,
                scale=scale,
                entity_alpha=entity_alpha,
            )
            self._draw_players(ctx=ctx, alive=False)
            self._draw_creatures(ctx=ctx)
            self._draw_freeze_overlay(ctx=ctx)
            self._draw_players(ctx=ctx, alive=True)
            self._draw_projectiles_and_effects(ctx=ctx)
            self._draw_bonus_and_ui(ctx=ctx, draw_aim_indicators=draw_aim_indicators)

    def _compute_view_transform(self) -> tuple[Vec2, Vec2, float, Vec2, Vec2]:
        out_w = float(rl.get_screen_width())
        out_h = float(rl.get_screen_height())
        out_size = Vec2(out_w, out_h)
        screen_size = self._camera_screen_size(runtime_w=out_w, runtime_h=out_h)
        camera = self._clamp_camera(self.camera, screen_size)
        scale_x = out_w / screen_size.x if screen_size.x > 0 else 1.0
        scale_y = out_h / screen_size.y if screen_size.y > 0 else 1.0
        view_scale = Vec2(scale_x, scale_y)
        scale = self._view_scale_avg(view_scale)
        return camera, view_scale, scale, screen_size, out_size

    def _draw_background(self, *, camera: Vec2, screen_size: Vec2, out_size: Vec2, view_scale: Vec2) -> None:
        clear_color = rl.Color(10, 10, 12, 255)
        ground = self.ground
        rl.clear_background(clear_color)
        if ground is not None:
            ground.draw(
                camera,
                screen_w=screen_size.x,
                screen_h=screen_size.y,
                out_w=out_size.x,
                out_h=out_size.y,
            )
            return

        # World bounds for debug if terrain is missing.
        world_min = camera.mul_components(view_scale)
        world_max = (camera + Vec2(float(self.world_size), float(self.world_size))).mul_components(view_scale)
        rl.draw_rectangle_lines(
            int(world_min.x),
            int(world_min.y),
            int(world_max.x - world_min.x),
            int(world_max.y - world_min.y),
            rl.Color(40, 40, 55, 255),
        )

    def _alpha_test_enabled(self) -> bool:
        ground = self.ground
        if ground is None:
            return True
        return bool(getattr(ground, "alpha_test", True))

    @staticmethod
    def _effect_src_rect(texture: rl.Texture, effect_id: EffectId) -> rl.Rectangle | None:
        atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(effect_id))
        if atlas is None:
            return None
        grid = SIZE_CODE_GRID.get(int(atlas.size_code))
        if not grid:
            return None
        frame = int(atlas.frame)
        col = frame % grid
        row = frame // grid
        cell_w = float(texture.width) / float(grid)
        cell_h = float(texture.height) / float(grid)
        return rl.Rectangle(
            cell_w * float(col),
            cell_h * float(row),
            max(0.0, cell_w - 2.0),
            max(0.0, cell_h - 2.0),
        )

    def _build_draw_context(
        self,
        *,
        camera: Vec2,
        view_scale: Vec2,
        scale: float,
        entity_alpha: float,
    ) -> WorldDrawContext:
        trooper_asset = CREATURE_ASSET.get(CreatureTypeId.TROOPER)
        trooper_texture = self.creature_textures.get(trooper_asset) if trooper_asset is not None else None
        particles_texture = self.particles_texture
        monster_vision = bool(self.players) and perk_active(self.players[0], PerkId.MONSTER_VISION)
        monster_vision_src = None
        if monster_vision and particles_texture is not None:
            monster_vision_src = self._effect_src_rect(particles_texture, EffectId.AURA)
        poison_src = None
        if particles_texture is not None:
            # Native uses `effect_select_texture(0x10)` (EffectId.AURA) for creature overlays
            # (monster vision, shadow, poison aura).
            poison_src = self._effect_src_rect(particles_texture, EffectId.AURA)
        return WorldDrawContext(
            camera=camera,
            view_scale=view_scale,
            scale=scale,
            entity_alpha=entity_alpha,
            trooper_texture=trooper_texture,
            particles_texture=particles_texture,
            monster_vision=monster_vision,
            monster_vision_src=monster_vision_src,
            poison_src=poison_src,
        )

    def _draw_player(self, player: PlayerState, *, ctx: WorldDrawContext) -> None:
        if ctx.trooper_texture is not None:
            self._draw_player_trooper_sprite(
                ctx.trooper_texture,
                player,
                camera=ctx.camera,
                view_scale=ctx.view_scale,
                scale=ctx.scale,
                alpha=ctx.entity_alpha,
            )
            return
        screen = self._world_to_screen_with(player.pos, camera=ctx.camera, view_scale=ctx.view_scale)
        tint = rl.Color(90, 190, 120, int(255 * ctx.entity_alpha + 0.5))
        rl.draw_circle(int(screen.x), int(screen.y), max(1.0, 14.0 * ctx.scale), tint)

    def _draw_players(self, *, ctx: WorldDrawContext, alive: bool) -> None:
        for player in self.players:
            if alive and player.health <= 0.0:
                continue
            if not alive and player.health > 0.0:
                continue
            self._draw_player(player, ctx=ctx)

    def _sorted_active_creatures(self) -> list[tuple[int, _CreatureDrawContract]]:
        creature_type_order = {
            int(CreatureTypeId.ZOMBIE): 0,
            int(CreatureTypeId.SPIDER_SP1): 1,
            int(CreatureTypeId.SPIDER_SP2): 2,
            int(CreatureTypeId.ALIEN): 3,
            int(CreatureTypeId.LIZARD): 4,
        }
        creatures = [(idx, creature) for idx, creature in enumerate(self.creatures.entries) if creature.active]
        creatures.sort(
            key=lambda item: (creature_type_order.get(int(getattr(item[1], "type_id", -1)), 999), item[0]),
        )
        return creatures

    def _draw_creature_overlays(
        self,
        creature: _CreatureDrawContract,
        *,
        screen: Vec2,
        hitbox_size: float,
        ctx: WorldDrawContext,
    ) -> None:
        fade = monster_vision_fade_alpha(hitbox_size)
        if ctx.monster_vision and ctx.particles_texture is not None and ctx.monster_vision_src is not None:
            mv_alpha = fade * ctx.entity_alpha
            if mv_alpha > 1e-3:
                size = 90.0 * ctx.scale
                dst = rl.Rectangle(screen.x, screen.y, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                tint = rl.Color(255, 255, 0, int(clamp(mv_alpha, 0.0, 1.0) * 255.0 + 0.5))
                rl.draw_texture_pro(ctx.particles_texture, ctx.monster_vision_src, dst, origin, 0.0, tint)
        if ctx.particles_texture is not None and ctx.poison_src is not None and bool(getattr(creature, "plague_infected", False)):
            # creature_render_all: collision_flag overlay (black 80x80 aura), drawn before red poison flag.
            plague_alpha = fade * ctx.entity_alpha
            if plague_alpha > 1e-3:
                size = 80.0 * ctx.scale
                dst = rl.Rectangle(screen.x, screen.y, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                tint = rl.Color(0, 0, 0, int(clamp(plague_alpha, 0.0, 1.0) * 255.0 + 0.5))
                rl.draw_texture_pro(ctx.particles_texture, ctx.poison_src, dst, origin, 0.0, tint)
        if (
            ctx.particles_texture is not None
            and ctx.poison_src is not None
            and (creature.flags & CreatureFlags.SELF_DAMAGE_TICK)
        ):
            poison_alpha = fade * ctx.entity_alpha
            if poison_alpha > 1e-3:
                size = 60.0 * ctx.scale
                dst = rl.Rectangle(screen.x, screen.y, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                tint = rl.Color(255, 0, 0, int(clamp(poison_alpha, 0.0, 1.0) * 255.0 + 0.5))
                rl.draw_texture_pro(ctx.particles_texture, ctx.poison_src, dst, origin, 0.0, tint)

    def _draw_creatures(self, *, ctx: WorldDrawContext) -> None:
        for _idx, creature in self._sorted_active_creatures():
            screen = self._world_to_screen_with(creature.pos, camera=ctx.camera, view_scale=ctx.view_scale)
            hitbox_size = float(creature.hitbox_size)
            try:
                type_id = CreatureTypeId(int(creature.type_id))
            except ValueError:
                type_id = None
            asset = CREATURE_ASSET.get(type_id) if type_id is not None else None
            texture = self.creature_textures.get(asset) if asset is not None else None
            self._draw_creature_overlays(creature, screen=screen, hitbox_size=hitbox_size, ctx=ctx)
            if texture is None:
                tint = rl.Color(220, 90, 90, int(255 * ctx.entity_alpha + 0.5))
                rl.draw_circle(int(screen.x), int(screen.y), max(1.0, creature.size * 0.5 * ctx.scale), tint)
                continue

            info = CREATURE_ANIM.get(type_id) if type_id is not None else None
            if info is None:
                continue

            tint_rgba = creature.tint

            # Energizer: tint "weak" creatures blue-ish while active.
            # Mirrors `creature_render_type` (0x00418b60) branch when
            # `_bonus_energizer_timer > 0` and `max_health < 500`.
            energizer_timer = float(self.state.bonuses.energizer)
            if energizer_timer > 0.0 and float(getattr(creature, "max_hp", 0.0)) < 500.0:
                # Native clamps to 1.0, then blends towards (0.5, 0.5, 1.0, 1.0).
                # Effect is full strength while timer >= 1 and fades out during the last second.
                t = energizer_timer
                if t >= 1.0:
                    t = 1.0
                elif t < 0.0:
                    t = 0.0
                tint_rgba = RGBA.lerp(tint_rgba, RGBA(0.5, 0.5, 1.0, 1.0), t)
            if hitbox_size < 0.0:
                # Mirrors the main-pass alpha fade when hitbox_size ramps negative.
                tint_rgba = tint_rgba.with_alpha(max(0.0, tint_rgba.a + hitbox_size * 0.1))
            tint = tint_rgba.scaled_alpha(ctx.entity_alpha).clamped().to_rl()

            size_scale = clamp(float(creature.size) / 64.0, 0.25, 2.0)
            fx_detail = self.config.fx_detail(level=0, default=True) if self.config is not None else True
            # Mirrors `creature_render_type`: the "shadow-ish" pass is gated by fx_detail_0
            # and is disabled when the Monster Vision perk is active.
            shadow = fx_detail and (not self.players or not perk_active(self.players[0], PerkId.MONSTER_VISION))
            long_strip = (creature.flags & CreatureFlags.ANIM_PING_PONG) == 0 or (
                creature.flags & CreatureFlags.ANIM_LONG_STRIP
            ) != 0
            phase = float(creature.anim_phase)
            if long_strip:
                if hitbox_size < 0.0:
                    # Negative phase selects the fallback "corpse" frame in creature_render_type.
                    phase = -1.0
                elif hitbox_size < 16.0:
                    # Death staging: while hitbox_size ramps down (16..0), creature_render_type
                    # selects frames via `__ftol((base_frame + 15) - hitbox_size)`.
                    phase = float(info.base + 0x0F) - hitbox_size - 0.5

            shadow_alpha = None
            if shadow:
                # Shadow pass uses tint_a * 0.4 and fades much faster for corpses (hitbox_size < 0).
                shadow_a = float(creature.tint.a) * 0.4
                if hitbox_size < 0.0:
                    shadow_a += hitbox_size * (0.5 if long_strip else 0.1)
                    shadow_a = max(0.0, shadow_a)
                shadow_alpha = int(clamp(shadow_a * ctx.entity_alpha * 255.0, 0.0, 255.0) + 0.5)
            self._draw_creature_sprite(
                texture,
                type_id=type_id or CreatureTypeId.ZOMBIE,
                flags=creature.flags,
                phase=phase,
                mirror_long=bool(info.mirror) and hitbox_size >= 16.0,
                shadow_alpha=shadow_alpha,
                pos=creature.pos,
                screen_pos=screen,
                rotation_rad=float(creature.heading) - math.pi / 2.0,
                scale=ctx.scale,
                size_scale=size_scale,
                tint=tint,
                shadow=shadow,
            )

    def _draw_freeze_overlay(self, *, ctx: WorldDrawContext) -> None:
        if ctx.particles_texture is None:
            return
        freeze_timer = float(self.state.bonuses.freeze)
        if freeze_timer <= 0.0:
            return
        src = self._effect_src_rect(ctx.particles_texture, EffectId.FREEZE_SHATTER)
        if src is None:
            return

        fade = 1.0 if freeze_timer >= 1.0 else clamp(freeze_timer, 0.0, 1.0)
        freeze_alpha = clamp(fade * ctx.entity_alpha * 0.7, 0.0, 1.0)
        if freeze_alpha <= 1e-3:
            return
        tint = rl.Color(255, 255, 255, int(freeze_alpha * 255.0 + 0.5))
        rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)
        for idx, creature in enumerate(self.creatures.entries):
            if not creature.active:
                continue
            size = float(creature.size) * ctx.scale
            if size <= 1e-3:
                continue
            creature_screen = self._world_to_screen_with(
                creature.pos,
                camera=ctx.camera,
                view_scale=ctx.view_scale,
            )
            dst = rl.Rectangle(creature_screen.x, creature_screen.y, size, size)
            origin = rl.Vector2(size * 0.5, size * 0.5)
            rotation_deg = (float(idx) * 0.01 + float(creature.heading)) * _RAD_TO_DEG
            rl.draw_texture_pro(ctx.particles_texture, src, dst, origin, rotation_deg, tint)
        rl.end_blend_mode()

    def _draw_projectiles_and_effects(self, *, ctx: WorldDrawContext) -> None:
        self._draw_sharpshooter_laser_sight(
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )
        for proj_index, proj in enumerate(self.state.projectiles.entries):
            if not proj.active:
                continue
            self._draw_projectile(proj, proj_index=proj_index, scale=ctx.scale, alpha=ctx.entity_alpha)
        self._draw_particle_pool(camera=ctx.camera, view_scale=ctx.view_scale, alpha=ctx.entity_alpha)
        for proj in self.state.secondary_projectiles.entries:
            if not proj.active:
                continue
            self._draw_secondary_projectile(proj, scale=ctx.scale, alpha=ctx.entity_alpha)
        self._draw_sprite_effect_pool(camera=ctx.camera, view_scale=ctx.view_scale, alpha=ctx.entity_alpha)
        self._draw_effect_pool(camera=ctx.camera, view_scale=ctx.view_scale, alpha=ctx.entity_alpha)

    def _draw_aim_indicators(self, *, ctx: WorldDrawContext) -> None:
        for player in self._iter_visible_aim_players():
            if player.health <= 0.0:
                continue
            aim = player.aim
            dist = player.pos.distance_to(player.aim)
            radius = max(6.0, dist * float(getattr(player, "spread_heat", 0.0)) * 0.5)
            aim_screen = self._world_to_screen_with(aim, camera=ctx.camera, view_scale=ctx.view_scale)
            screen_radius = max(1.0, radius * ctx.scale)
            self._draw_aim_circle(center=aim_screen, radius=screen_radius, alpha=ctx.entity_alpha)
            reload_timer = float(getattr(player, "reload_timer", 0.0))
            reload_max = float(getattr(player, "reload_timer_max", 0.0))
            if reload_max > 1e-6 and reload_timer > 1e-6:
                progress = reload_timer / reload_max
                if progress > 0.0:
                    ms = int(progress * 60000.0)
                    self._draw_clock_gauge(
                        pos=Vec2(int(aim_screen.x), int(aim_screen.y)),
                        ms=ms,
                        scale=ctx.scale,
                        alpha=ctx.entity_alpha,
                    )

    def _draw_aim_enhancements(self, *, ctx: WorldDrawContext) -> None:
        for player in self._iter_visible_aim_players():
            if player.health <= 0.0:
                continue
            aim_screen = self._world_to_screen_with(player.aim, camera=ctx.camera, view_scale=ctx.view_scale)
            draw_aim_cursor(ctx.particles_texture, self.aim_texture, pos=aim_screen)

    def _iter_visible_aim_players(self) -> tuple[PlayerState, ...]:
        players = self.players
        if not bool(self.lan_local_aim_indicators_only):
            return tuple(players)
        local_slot = int(self.lan_local_player_slot_index)
        return tuple(player for player in players if int(player.index) == local_slot)

    def _draw_bonus_and_ui(self, *, ctx: WorldDrawContext, draw_aim_indicators: bool) -> None:
        self._draw_bonus_pickups(
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )
        self._draw_bonus_hover_labels(camera=ctx.camera, view_scale=ctx.view_scale, alpha=ctx.entity_alpha)
        draw_world_aim = draw_aim_indicators and (not self.demo_mode_active)
        if draw_world_aim:
            self._draw_aim_indicators(ctx=ctx)
        self._draw_direction_arrows(
            camera=ctx.camera,
            view_scale=ctx.view_scale,
            scale=ctx.scale,
            alpha=ctx.entity_alpha,
        )
        if draw_world_aim:
            self._draw_aim_enhancements(ctx=ctx)



from __future__ import annotations

from dataclasses import dataclass

import pyray as rl

from grim.fonts.small import SmallFontData, load_small_font
from grim.geom import Vec2
from grim.view import View, ViewContext

from ..weapons import WEAPON_TABLE, Weapon, weapon_display_name
from ._ui_helpers import draw_ui_text, ui_line_height
from .registry import register_view

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)
UI_HOVER_COLOR = rl.Color(240, 200, 80, 255)


@dataclass(frozen=True, slots=True)
class WeaponIconGroup:
    icon_index: int
    weapons: tuple[Weapon, ...]


def _build_icon_groups() -> dict[int, WeaponIconGroup]:
    grouped: dict[int, list[Weapon]] = {}
    for entry in WEAPON_TABLE:
        icon_index = entry.icon_index
        if icon_index is None or icon_index < 0 or icon_index > 31:
            continue
        grouped.setdefault(icon_index, []).append(entry)
    return {
        icon_index: WeaponIconGroup(icon_index=icon_index, weapons=tuple(entries))
        for icon_index, entries in grouped.items()
    }


WEAPON_ICON_GROUPS = _build_icon_groups()


class WeaponIconView:
    def __init__(self, ctx: ViewContext) -> None:
        self._assets_root = ctx.assets_dir
        self._preserve_bugs = bool(ctx.preserve_bugs)
        self._missing_assets: list[str] = []
        self._texture: rl.Texture | None = None
        self._small: SmallFontData | None = None

    def open(self) -> None:
        self._missing_assets.clear()
        self._small = load_small_font(self._assets_root)
        path = self._assets_root / "crimson" / "ui" / "ui_wicons.png"
        if not path.is_file():
            self._missing_assets.append("ui/ui_wicons.png")
            raise FileNotFoundError(f"Missing asset: {path}")
        self._texture = rl.load_texture(str(path))

    def close(self) -> None:
        if self._texture is not None:
            rl.unload_texture(self._texture)
            self._texture = None
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None

    def update(self, dt: float) -> None:
        del dt

    def draw(self) -> None:
        rl.clear_background(rl.Color(12, 12, 14, 255))
        if self._missing_assets:
            message = "Missing assets: " + ", ".join(self._missing_assets)
            draw_ui_text(self._small, message, Vec2(24, 24), scale=UI_TEXT_SCALE, color=UI_ERROR_COLOR)
            return
        if self._texture is None:
            draw_ui_text(
                self._small,
                "No weapon icon texture loaded.",
                Vec2(24, 24),
                scale=UI_TEXT_SCALE,
                color=UI_TEXT_COLOR,
            )
            return

        margin = 24
        panel_gap = 32
        panel_width = min(420, int(rl.get_screen_width() * 0.4))
        available_width = rl.get_screen_width() - margin * 2 - panel_gap - panel_width
        available_height = rl.get_screen_height() - margin * 2 - 60

        cols = 4
        rows = 8
        icon_w = self._texture.width / cols
        icon_h = self._texture.height / rows
        scale = min(2.5, available_width / (cols * icon_w), available_height / (rows * icon_h))

        x = margin
        y = margin + 60
        hovered_index = None
        mouse = rl.get_mouse_position()

        for idx in range(cols * rows):
            row = idx // cols
            col = idx % cols
            dst_x = x + col * icon_w * scale
            dst_y = y + row * icon_h * scale
            dst = rl.Rectangle(float(dst_x), float(dst_y), float(icon_w * scale), float(icon_h * scale))
            src = rl.Rectangle(float(col * icon_w), float(row * icon_h), float(icon_w), float(icon_h))
            rl.draw_texture_pro(self._texture, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

            if dst_x <= mouse.x <= dst_x + dst.width and dst_y <= mouse.y <= dst_y + dst.height:
                hovered_index = idx
                rl.draw_rectangle_lines_ex(dst, 3, UI_HOVER_COLOR)

            draw_ui_text(
                self._small,
                f"{idx:02d}",
                Vec2(dst_x + 4, dst_y + 4),
                scale=0.75,
                color=UI_HINT_COLOR,
            )

        info_x = x + cols * icon_w * scale + panel_gap
        info_y = margin
        draw_ui_text(
            self._small,
            "ui_wicons.png (8x8 grid, 2x1 subrects)",
            Vec2(info_x, info_y),
            scale=UI_TEXT_SCALE,
            color=UI_TEXT_COLOR,
        )
        info_y += ui_line_height(self._small, scale=UI_TEXT_SCALE) + 12

        if hovered_index is not None:
            frame = hovered_index * 2
            draw_ui_text(
                self._small,
                f"icon_index {hovered_index}  frame {frame}",
                Vec2(info_x, info_y),
                scale=UI_TEXT_SCALE,
                color=UI_TEXT_COLOR,
            )
            info_y += ui_line_height(self._small, scale=UI_TEXT_SCALE) + 6
            group = WEAPON_ICON_GROUPS.get(hovered_index)
            if group is None:
                draw_ui_text(
                    self._small, "no weapon mapping", Vec2(info_x, info_y), scale=UI_TEXT_SCALE, color=UI_HINT_COLOR,
                )
                info_y += ui_line_height(self._small, scale=UI_TEXT_SCALE) + 6
            else:
                for weapon in group.weapons:
                    name = weapon_display_name(int(weapon.weapon_id), preserve_bugs=bool(self._preserve_bugs))
                    draw_ui_text(self._small, name, Vec2(info_x, info_y), scale=UI_TEXT_SCALE, color=UI_TEXT_COLOR)
                    info_y += ui_line_height(self._small, scale=UI_TEXT_SCALE) + 4


@register_view("wicons", "Weapon icon preview")
def build_weapon_icon_view(ctx: ViewContext) -> View:
    return WeaponIconView(ctx)



from __future__ import annotations

"""
PAQ archive format (Crimsonland).

File layout:
  - magic: 4 bytes, ASCII "paq\\0"
  - entries: repeated until EOF
      - name: NUL-terminated UTF-8 string (relative path)
      - size: u32 little-endian payload size
      - payload: raw file bytes of length `size`
"""

from collections.abc import Iterable, Iterator
from pathlib import Path

from construct import Bytes, Const, CString, GreedyRange, Int32ul, Struct

MAGIC = b"paq\x00"


PAQ_ENTRY = Struct(
    "name" / CString("utf8"),
    "size" / Int32ul,
    "payload" / Bytes(lambda ctx: ctx.size),
)

PAQ = Struct(
    "magic" / Const(MAGIC),
    "entries" / GreedyRange(PAQ_ENTRY),
)


def iter_entries_bytes(data: bytes) -> Iterator[tuple[str, bytes]]:
    parsed = PAQ.parse(data)
    for entry in parsed.entries:
        yield entry.name, entry.payload


def iter_entries(source: str | Path) -> Iterator[tuple[str, bytes]]:
    data = Path(source).read_bytes()
    yield from iter_entries_bytes(data)


def read_paq(source: str | Path) -> list[tuple[str, bytes]]:
    return list(iter_entries(source))


def decode_bytes(data: bytes) -> list[tuple[str, bytes]]:
    return list(iter_entries_bytes(data))


def build_entries(entries: Iterable[tuple[str, bytes]]) -> bytes:
    built_entries = []
    for name, data in entries:
        if isinstance(name, Path):
            name = str(name)
        if isinstance(data, memoryview):
            data = data.tobytes()
        built_entries.append(
            {
                "name": str(name),
                "size": len(data),
                "payload": bytes(data),
            },
        )
    return PAQ.build({"magic": MAGIC, "entries": built_entries})


def write_paq(dest: str | Path, entries: Iterable[tuple[str, bytes]]) -> None:
    data = build_entries(entries)
    Path(dest).write_bytes(data)


def encode_bytes(entries: Iterable[tuple[str, bytes]]) -> bytes:
    return build_entries(entries)



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass

from ...sim.state_types import GameplayState, PlayerState
from ..ids import PerkId
from ..state import CreatureForPerks, PerkSelectionState


@dataclass(slots=True)
class PerkApplyCtx:
    state: GameplayState
    players: list[PlayerState]
    owner: PlayerState
    perk_id: PerkId
    perk_state: PerkSelectionState | None
    dt: float | None
    creatures: Sequence[CreatureForPerks] | None

    def frame_dt(self) -> float:
        return float(self.dt) if self.dt is not None else 0.0


PerkApplyHandler = Callable[[PerkApplyCtx], None]



from __future__ import annotations

from dataclasses import replace
from typing import cast

from ...game_modes import GameMode
from ...original.capture import CAPTURE_BOOTSTRAP_EVENT_KIND
from ...quests import quest_by_level
from ...quests.runtime import build_quest_spawn_table
from ...quests.types import QuestContext, SpawnEntry
from ...replay import Replay, UnknownEvent, apply_replay_bootstrap, unpack_tick_inputs, warn_on_game_version_mismatch
from ...replay.checkpoints import ReplayCheckpoint, build_checkpoint
from ...weapon_runtime import weapon_assign_player
from ...weapons import WeaponId
from ..sessions import QuestDeterministicSession, RushDeterministicSession, SurvivalDeterministicSession
from ..world_state import WorldEvents, WorldState
from .replay_events import apply_replay_tick_events, partition_tick_events
from .replay_timing import resolve_dt_frame, resolve_dt_frame_ms_i32, should_apply_world_dt_steps_for_replay
from .setup import (
    ReplayRunnerError,
    RunResult,
    build_damage_scale_by_type,
    build_empty_fx_queues,
    player0_most_used_weapon_id,
    player0_shots,
    reset_players,
    status_from_snapshot,
)

RUSH_WEAPON_ID = WeaponId.ASSAULT_RIFLE


def _enforce_rush_loadout(world: WorldState) -> None:
    for player in world.players:
        if int(player.weapon_id) != int(RUSH_WEAPON_ID):
            weapon_assign_player(player, int(RUSH_WEAPON_ID))
        # `rush_mode_update` forces weapon+ammo every frame; keep ammo topped up.
        player.ammo = float(max(0, int(player.clip_size)))


def run_survival_replay(
    replay: Replay,
    *,
    max_ticks: int | None = None,
    warn_on_version_mismatch: bool = True,
    strict_events: bool = True,
    trace_rng: bool = False,
    checkpoint_use_world_step_creature_count: bool = False,
    checkpoints_out: list[ReplayCheckpoint] | None = None,
    checkpoint_ticks: set[int] | None = None,
    dt_frame_overrides: dict[int, float] | None = None,
    dt_frame_ms_i32_overrides: dict[int, int] | None = None,
    inter_tick_rand_draws: int = 0,
    inter_tick_rand_draws_by_tick: dict[int, int] | None = None,
) -> RunResult:
    if int(replay.header.game_mode_id) != int(GameMode.SURVIVAL):
        raise ReplayRunnerError(
            f"replay game_mode_id={int(replay.header.game_mode_id)} does not match survival={int(GameMode.SURVIVAL)}",
        )

    if warn_on_version_mismatch:
        warn_on_game_version_mismatch(replay, action="verification")

    tick_rate = int(replay.header.tick_rate)
    if tick_rate <= 0:
        raise ReplayRunnerError(f"invalid tick_rate: {tick_rate}")
    dt_frame = 1.0 / float(tick_rate)

    world_size = float(replay.header.world_size)
    world = WorldState.build(
        world_size=world_size,
        demo_mode_active=False,
        hardcore=bool(replay.header.hardcore),
        difficulty_level=int(replay.header.difficulty_level),
        preserve_bugs=bool(replay.header.preserve_bugs),
    )
    reset_players(
        world.players,
        world_size=world_size,
        player_count=int(replay.header.player_count),
    )
    world.state.status = status_from_snapshot(
        quest_unlock_index=int(replay.header.status.quest_unlock_index),
        quest_unlock_index_full=int(replay.header.status.quest_unlock_index_full),
        weapon_usage_counts=replay.header.status.weapon_usage_counts,
    )
    apply_replay_bootstrap(
        replay.header,
        rng=world.state.rng,
        world_size=float(world_size),
        strict=True,
    )

    events_by_tick: dict[int, list[object]] = {}
    original_capture_replay = False
    for event in replay.events:
        if isinstance(event, UnknownEvent) and str(event.kind) == CAPTURE_BOOTSTRAP_EVENT_KIND:
            original_capture_replay = True
        events_by_tick.setdefault(int(event.tick_index), []).append(event)
    apply_world_dt_steps = should_apply_world_dt_steps_for_replay(
        original_capture_replay=bool(original_capture_replay),
        dt_frame_overrides=dt_frame_overrides,
        dt_frame_ms_i32_overrides=dt_frame_ms_i32_overrides,
    )

    fx_queue, fx_queue_rotated = build_empty_fx_queues()
    damage_scale_by_type = build_damage_scale_by_type()
    session = SurvivalDeterministicSession(
        world=world,
        world_size=float(world_size),
        damage_scale_by_type=damage_scale_by_type,
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        detail_preset=int(replay.header.detail_preset),
        fx_toggle=int(replay.header.fx_toggle),
        game_tune_started=False,
        apply_world_dt_steps=bool(apply_world_dt_steps),
        clear_fx_queues_each_tick=True,
    )

    inputs = replay.inputs
    tick_limit = len(inputs) if max_ticks is None else min(len(inputs), max(0, int(max_ticks)))

    for tick_index in range(tick_limit):
        state = world.state
        state.game_mode = int(GameMode.SURVIVAL)
        state.demo_mode_active = False
        if inter_tick_rand_draws_by_tick is not None:
            draws = inter_tick_rand_draws_by_tick.get(int(tick_index))
            if draws is None:
                draws = int(inter_tick_rand_draws)
            for _ in range(max(0, int(draws))):
                world.state.rng.rand()

        dt_tick = resolve_dt_frame(
            tick_index=int(tick_index),
            default_dt_frame=float(dt_frame),
            dt_frame_overrides=dt_frame_overrides,
        )
        dt_tick_ms_i32 = resolve_dt_frame_ms_i32(
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            dt_frame_ms_i32_overrides=dt_frame_ms_i32_overrides,
        )

        tick_events = events_by_tick.get(tick_index, [])
        pre_step_events, post_step_events = partition_tick_events(
            tick_events,
            defer_menu_open=bool(original_capture_replay),
        )

        rng_before_events = int(state.rng.state)
        apply_replay_tick_events(
            pre_step_events,
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            world=world,
            game_mode_id=int(GameMode.SURVIVAL),
            strict_events=bool(strict_events),
        )
        rng_after_events = int(state.rng.state)

        player_inputs = unpack_tick_inputs(inputs[tick_index])

        tick = session.step_tick(
            dt_frame=float(dt_tick),
            dt_frame_ms_i32=(int(dt_tick_ms_i32) if dt_tick_ms_i32 is not None else None),
            inputs=player_inputs,
            trace_rng=bool(trace_rng),
        )
        step = tick.step
        events = step.events

        rng_before_post_events = int(state.rng.state)
        if post_step_events:
            apply_replay_tick_events(
                post_step_events,
                tick_index=int(tick_index),
                dt_frame=float(dt_tick),
                world=world,
                game_mode_id=int(GameMode.SURVIVAL),
                strict_events=bool(strict_events),
            )
        rng_after_post_events = int(state.rng.state)

        if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
            checkpoint_rng_marks = dict(tick.rng_marks)
            checkpoint_rng_marks["before_events"] = int(rng_before_events)
            checkpoint_rng_marks["after_events"] = (
                int(rng_after_post_events) if post_step_events else int(rng_after_events)
            )
            if post_step_events:
                checkpoint_rng_marks["before_post_events"] = int(rng_before_post_events)
                checkpoint_rng_marks["after_post_events"] = int(rng_after_post_events)
            checkpoints_out.append(
                build_checkpoint(
                    tick_index=int(tick_index),
                    world=world,
                    elapsed_ms=float(tick.elapsed_ms),
                    creature_count_override=(
                        int(tick.creature_count_world_step) if checkpoint_use_world_step_creature_count else None
                    ),
                    rng_marks=checkpoint_rng_marks,
                    deaths=events.deaths,
                    events=events,
                    command_hash=str(step.command_hash),
                ),
            )

        if inter_tick_rand_draws_by_tick is None:
            draws = max(0, int(inter_tick_rand_draws))
            for _ in range(draws):
                world.state.rng.rand()
    else:
        tick_index = tick_limit

    # Some UI-side events (e.g. final perk picks) can be recorded at the
    # terminal boundary tick == len(inputs), after the last simulated input.
    if int(tick_index) == int(len(inputs)):
        dt_tick = resolve_dt_frame(
            tick_index=int(tick_index),
            default_dt_frame=float(dt_frame),
            dt_frame_overrides=dt_frame_overrides,
        )
        apply_replay_tick_events(
            events_by_tick.get(int(tick_index), []),
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            world=world,
            game_mode_id=int(GameMode.SURVIVAL),
            strict_events=bool(strict_events),
        )
        if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
            checkpoints_out.append(
                build_checkpoint(
                    tick_index=int(tick_index),
                    world=world,
                    elapsed_ms=float(session.elapsed_ms),
                    rng_marks={},
                    deaths=[],
                    events=WorldEvents(hits=[], deaths=(), pickups=[], sfx=[]),
                    command_hash="",
                ),
            )

    shots_fired, shots_hit = player0_shots(world.state)
    most_used_weapon_id = player0_most_used_weapon_id(world.state, world.players)
    score_xp = int(world.players[0].experience) if world.players else 0

    return RunResult(
        game_mode_id=int(GameMode.SURVIVAL),
        tick_rate=tick_rate,
        ticks=int(tick_index),
        elapsed_ms=int(session.elapsed_ms),
        score_xp=score_xp,
        creature_kill_count=int(world.creatures.kill_count),
        most_used_weapon_id=int(most_used_weapon_id),
        shots_fired=int(shots_fired),
        shots_hit=int(shots_hit),
        rng_state=int(world.state.rng.state),
    )


def run_rush_replay(
    replay: Replay,
    *,
    max_ticks: int | None = None,
    warn_on_version_mismatch: bool = True,
    trace_rng: bool = False,
    checkpoint_use_world_step_creature_count: bool = False,
    checkpoints_out: list[ReplayCheckpoint] | None = None,
    checkpoint_ticks: set[int] | None = None,
    dt_frame_overrides: dict[int, float] | None = None,
    inter_tick_rand_draws: int = 0,
    inter_tick_rand_draws_by_tick: dict[int, int] | None = None,
) -> RunResult:
    if int(replay.header.game_mode_id) != int(GameMode.RUSH):
        raise ReplayRunnerError(
            f"replay game_mode_id={int(replay.header.game_mode_id)} does not match rush={int(GameMode.RUSH)}",
        )

    if warn_on_version_mismatch:
        warn_on_game_version_mismatch(replay, action="verification")

    events_by_tick: dict[int, list[UnknownEvent]] = {}
    for event in replay.events:
        if isinstance(event, UnknownEvent) and str(event.kind) == CAPTURE_BOOTSTRAP_EVENT_KIND:
            events_by_tick.setdefault(int(event.tick_index), []).append(event)
            continue
        raise ReplayRunnerError("rush replay does not support events")

    tick_rate = int(replay.header.tick_rate)
    if tick_rate <= 0:
        raise ReplayRunnerError(f"invalid tick_rate: {tick_rate}")
    dt_frame = 1.0 / float(tick_rate)

    world_size = float(replay.header.world_size)
    world = WorldState.build(
        world_size=world_size,
        demo_mode_active=False,
        hardcore=bool(replay.header.hardcore),
        difficulty_level=int(replay.header.difficulty_level),
        preserve_bugs=bool(replay.header.preserve_bugs),
    )
    reset_players(
        world.players,
        world_size=world_size,
        player_count=int(replay.header.player_count),
    )
    world.state.status = status_from_snapshot(
        quest_unlock_index=int(replay.header.status.quest_unlock_index),
        quest_unlock_index_full=int(replay.header.status.quest_unlock_index_full),
        weapon_usage_counts=replay.header.status.weapon_usage_counts,
    )
    apply_replay_bootstrap(
        replay.header,
        rng=world.state.rng,
        world_size=float(world_size),
        strict=True,
    )

    _enforce_rush_loadout(world)

    fx_queue, fx_queue_rotated = build_empty_fx_queues()
    damage_scale_by_type = build_damage_scale_by_type()
    session = RushDeterministicSession(
        world=world,
        world_size=float(world_size),
        damage_scale_by_type=damage_scale_by_type,
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        detail_preset=int(replay.header.detail_preset),
        fx_toggle=int(replay.header.fx_toggle),
        game_tune_started=False,
        clear_fx_queues_each_tick=True,
        enforce_loadout=lambda: _enforce_rush_loadout(world),
    )

    inputs = replay.inputs
    tick_limit = len(inputs) if max_ticks is None else min(len(inputs), max(0, int(max_ticks)))

    for tick_index in range(tick_limit):
        state = world.state
        state.game_mode = int(GameMode.RUSH)
        state.demo_mode_active = False
        if inter_tick_rand_draws_by_tick is not None:
            draws = inter_tick_rand_draws_by_tick.get(int(tick_index))
            if draws is None:
                draws = int(inter_tick_rand_draws)
            for _ in range(max(0, int(draws))):
                world.state.rng.rand()
        dt_tick = resolve_dt_frame(
            tick_index=int(tick_index),
            default_dt_frame=float(dt_frame),
            dt_frame_overrides=dt_frame_overrides,
        )

        rng_before_events = int(state.rng.state)
        apply_replay_tick_events(
            cast("list[object]", events_by_tick.get(int(tick_index), [])),
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            world=world,
            game_mode_id=int(GameMode.RUSH),
            strict_events=True,
        )
        rng_after_events = int(state.rng.state)

        player_inputs = [replace(inp, reload_pressed=False) for inp in unpack_tick_inputs(inputs[tick_index])]

        tick = session.step_tick(
            dt_frame=float(dt_tick),
            inputs=player_inputs,
            trace_rng=bool(trace_rng),
        )
        step = tick.step
        events = step.events

        if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
            checkpoint_rng_marks = dict(tick.rng_marks)
            checkpoint_rng_marks["before_events"] = int(rng_before_events)
            checkpoint_rng_marks["after_events"] = int(rng_after_events)
            checkpoints_out.append(
                build_checkpoint(
                    tick_index=int(tick_index),
                    world=world,
                    elapsed_ms=float(tick.elapsed_ms),
                    creature_count_override=(
                        int(tick.creature_count_world_step) if checkpoint_use_world_step_creature_count else None
                    ),
                    rng_marks=checkpoint_rng_marks,
                    deaths=events.deaths,
                    events=events,
                    command_hash=str(step.command_hash),
                ),
            )

        if inter_tick_rand_draws_by_tick is None:
            draws = max(0, int(inter_tick_rand_draws))
            for _ in range(draws):
                world.state.rng.rand()
    else:
        tick_index = tick_limit

    rng_before_events = int(world.state.rng.state)
    apply_replay_tick_events(
        cast("list[object]", events_by_tick.get(int(tick_index), [])),
        tick_index=int(tick_index),
        dt_frame=float(dt_frame),
        world=world,
        game_mode_id=int(GameMode.RUSH),
        strict_events=True,
    )
    rng_after_events = int(world.state.rng.state)

    if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
        checkpoints_out.append(
            build_checkpoint(
                tick_index=int(tick_index),
                world=world,
                elapsed_ms=float(session.elapsed_ms),
                rng_marks={"before_events": int(rng_before_events), "after_events": int(rng_after_events)},
                deaths=[],
                events=WorldEvents(hits=[], deaths=(), pickups=[], sfx=[]),
                command_hash="",
            ),
        )

    shots_fired, shots_hit = player0_shots(world.state)
    most_used_weapon_id = player0_most_used_weapon_id(world.state, world.players)
    score_xp = int(world.players[0].experience) if world.players else 0

    return RunResult(
        game_mode_id=int(GameMode.RUSH),
        tick_rate=tick_rate,
        ticks=int(tick_index),
        elapsed_ms=int(session.elapsed_ms),
        score_xp=score_xp,
        creature_kill_count=int(world.creatures.kill_count),
        most_used_weapon_id=int(most_used_weapon_id),
        shots_fired=int(shots_fired),
        shots_hit=int(shots_hit),
        rng_state=int(world.state.rng.state),
    )


def _resolve_quest_level(replay: Replay) -> str:
    quest_level = str(getattr(replay.header, "quest_level", "") or "")
    if quest_level:
        return str(quest_level)

    # Legacy replays (e.g. capture-derived) may not encode the quest id. Classic quest RNG
    # seeding uses `major*100 + minor`, so we can often recover the level from `header.seed`.
    seed = int(replay.header.seed)
    major = seed // 100
    minor = seed % 100
    if 1 <= int(major) <= 5 and 1 <= int(minor) <= 10:
        return f"{major}.{minor}"
    return ""


def run_quest_replay(
    replay: Replay,
    *,
    spawn_entries: tuple[SpawnEntry, ...] | None = None,
    quest_stage_major: int | None = None,
    quest_stage_minor: int | None = None,
    start_weapon_id: int | None = None,
    max_ticks: int | None = None,
    warn_on_version_mismatch: bool = True,
    strict_events: bool = True,
    trace_rng: bool = False,
    checkpoint_use_world_step_creature_count: bool = False,
    checkpoints_out: list[ReplayCheckpoint] | None = None,
    checkpoint_ticks: set[int] | None = None,
    dt_frame_overrides: dict[int, float] | None = None,
    inter_tick_rand_draws: int = 0,
    inter_tick_rand_draws_by_tick: dict[int, int] | None = None,
) -> RunResult:
    if int(replay.header.game_mode_id) != int(GameMode.QUESTS):
        raise ReplayRunnerError(
            f"replay game_mode_id={int(replay.header.game_mode_id)} does not match quests={int(GameMode.QUESTS)}",
        )

    if warn_on_version_mismatch:
        warn_on_game_version_mismatch(replay, action="verification")

    tick_rate = int(replay.header.tick_rate)
    if tick_rate <= 0:
        raise ReplayRunnerError(f"invalid tick_rate: {tick_rate}")
    dt_frame = 1.0 / float(tick_rate)

    world_size = float(replay.header.world_size)
    world = WorldState.build(
        world_size=world_size,
        demo_mode_active=False,
        hardcore=bool(replay.header.hardcore),
        difficulty_level=int(replay.header.difficulty_level),
        preserve_bugs=bool(replay.header.preserve_bugs),
    )
    reset_players(
        world.players,
        world_size=world_size,
        player_count=int(replay.header.player_count),
    )
    world.state.status = status_from_snapshot(
        quest_unlock_index=int(replay.header.status.quest_unlock_index),
        quest_unlock_index_full=int(replay.header.status.quest_unlock_index_full),
        weapon_usage_counts=replay.header.status.weapon_usage_counts,
    )
    world.state.rng.srand(int(replay.header.seed))

    if spawn_entries is None:
        quest_level = _resolve_quest_level(replay)
        quest = quest_by_level(quest_level) if quest_level else None
        if quest is None:
            raise ReplayRunnerError(f"unsupported quest replay: unknown quest_level={quest_level!r}")

        if quest_stage_major is None or quest_stage_minor is None:
            world.state.quest_stage_major, world.state.quest_stage_minor = quest.level_key
        if start_weapon_id is None:
            start_weapon_id = int(getattr(quest, "start_weapon_id", 1) or 1)

        ctx = QuestContext(
            width=int(world_size),
            height=int(world_size),
            player_count=int(replay.header.player_count),
        )
        spawn_entries = tuple(
            build_quest_spawn_table(
                quest,
                ctx,
                seed=int(replay.header.seed),
                hardcore=bool(replay.header.hardcore),
                full_version=True,
            ),
        )
    else:
        spawn_entries = tuple(spawn_entries)

    if quest_stage_major is not None:
        world.state.quest_stage_major = int(quest_stage_major)
    if quest_stage_minor is not None:
        world.state.quest_stage_minor = int(quest_stage_minor)

    weapon_id = max(1, int(start_weapon_id or 1))
    for player in world.players:
        weapon_assign_player(player, weapon_id)

    events_by_tick: dict[int, list[object]] = {}
    original_capture_replay = False
    bootstrap_start_tick: int | None = None
    for event in replay.events:
        events_by_tick.setdefault(int(event.tick_index), []).append(event)
        if isinstance(event, UnknownEvent) and str(event.kind) == CAPTURE_BOOTSTRAP_EVENT_KIND:
            original_capture_replay = True
            tick_index = int(event.tick_index)
            if bootstrap_start_tick is None or tick_index < int(bootstrap_start_tick):
                bootstrap_start_tick = int(tick_index)

    fx_queue, fx_queue_rotated = build_empty_fx_queues()
    damage_scale_by_type = build_damage_scale_by_type()
    session = QuestDeterministicSession(
        world=world,
        world_size=float(world_size),
        damage_scale_by_type=damage_scale_by_type,
        fx_queue=fx_queue,
        fx_queue_rotated=fx_queue_rotated,
        spawn_entries=tuple(spawn_entries),
        detail_preset=int(replay.header.detail_preset),
        fx_toggle=int(replay.header.fx_toggle),
        clear_fx_queues_each_tick=True,
    )

    inputs = replay.inputs
    tick_limit = len(inputs) if max_ticks is None else min(len(inputs), max(0, int(max_ticks)))
    tick_start = max(0, int(bootstrap_start_tick)) if bootstrap_start_tick is not None else 0

    for tick_index in range(int(tick_start), int(tick_limit)):
        state = world.state
        state.game_mode = int(GameMode.QUESTS)
        state.demo_mode_active = False

        if inter_tick_rand_draws_by_tick is not None:
            draws = inter_tick_rand_draws_by_tick.get(int(tick_index))
            if draws is None:
                draws = int(inter_tick_rand_draws)
            for _ in range(max(0, int(draws))):
                world.state.rng.rand()

        dt_tick = resolve_dt_frame(
            tick_index=int(tick_index),
            default_dt_frame=float(dt_frame),
            dt_frame_overrides=dt_frame_overrides,
        )

        tick_events = events_by_tick.get(int(tick_index), [])
        pre_step_events, post_step_events = partition_tick_events(
            tick_events,
            defer_menu_open=bool(original_capture_replay),
        )
        rng_before_events = int(state.rng.state)
        apply_replay_tick_events(
            pre_step_events,
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            world=world,
            game_mode_id=int(GameMode.QUESTS),
            strict_events=bool(strict_events),
        )
        rng_after_events = int(state.rng.state)

        player_inputs = unpack_tick_inputs(inputs[tick_index])

        tick = session.step_tick(
            dt_frame=float(dt_tick),
            inputs=player_inputs,
            trace_rng=bool(trace_rng),
        )
        step = tick.step
        events = step.events

        rng_before_post_events = int(state.rng.state)
        if post_step_events:
            apply_replay_tick_events(
                post_step_events,
                tick_index=int(tick_index),
                dt_frame=float(dt_tick),
                world=world,
                game_mode_id=int(GameMode.QUESTS),
                strict_events=bool(strict_events),
            )
        rng_after_post_events = int(state.rng.state)

        if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
            checkpoint_rng_marks = dict(tick.rng_marks)
            checkpoint_rng_marks["before_events"] = int(rng_before_events)
            checkpoint_rng_marks["after_events"] = (
                int(rng_after_post_events) if post_step_events else int(rng_after_events)
            )
            if post_step_events:
                checkpoint_rng_marks["before_post_events"] = int(rng_before_post_events)
                checkpoint_rng_marks["after_post_events"] = int(rng_after_post_events)
            checkpoints_out.append(
                build_checkpoint(
                    tick_index=int(tick_index),
                    world=world,
                    elapsed_ms=float(tick.spawn_timeline_ms),
                    creature_count_override=(
                        int(tick.creature_count_world_step) if checkpoint_use_world_step_creature_count else None
                    ),
                    rng_marks=checkpoint_rng_marks,
                    deaths=events.deaths,
                    events=events,
                    command_hash=str(step.command_hash),
                ),
            )

        if inter_tick_rand_draws_by_tick is None:
            draws = max(0, int(inter_tick_rand_draws))
            for _ in range(draws):
                world.state.rng.rand()

    else:
        tick_index = tick_limit

    if int(tick_index) == int(len(inputs)):
        dt_tick = resolve_dt_frame(
            tick_index=int(tick_index),
            default_dt_frame=float(dt_frame),
            dt_frame_overrides=dt_frame_overrides,
        )
        apply_replay_tick_events(
            events_by_tick.get(int(tick_index), []),
            tick_index=int(tick_index),
            dt_frame=float(dt_tick),
            world=world,
            game_mode_id=int(GameMode.QUESTS),
            strict_events=bool(strict_events),
        )
        if checkpoints_out is not None and checkpoint_ticks is not None and int(tick_index) in checkpoint_ticks:
            checkpoints_out.append(
                build_checkpoint(
                    tick_index=int(tick_index),
                    world=world,
                    elapsed_ms=float(session.spawn_timeline_ms),
                    rng_marks={},
                    deaths=[],
                    events=WorldEvents(hits=[], deaths=(), pickups=[], sfx=[]),
                    command_hash="",
                ),
            )

    shots_fired, shots_hit = player0_shots(world.state)
    most_used_weapon_id = player0_most_used_weapon_id(world.state, world.players)
    score_xp = int(world.players[0].experience) if world.players else 0

    return RunResult(
        game_mode_id=int(GameMode.QUESTS),
        tick_rate=tick_rate,
        ticks=int(tick_index),
        elapsed_ms=int(session.spawn_timeline_ms),
        score_xp=score_xp,
        creature_kill_count=int(world.creatures.kill_count),
        most_used_weapon_id=int(most_used_weapon_id),
        shots_fired=int(shots_fired),
        shots_hit=int(shots_hit),
        rng_state=int(world.state.rng.state),
    )


def run_replay(
    replay: Replay,
    *,
    max_ticks: int | None = None,
    warn_on_version_mismatch: bool = True,
    strict_events: bool = True,
    trace_rng: bool = False,
    checkpoint_use_world_step_creature_count: bool = False,
    checkpoints_out: list[ReplayCheckpoint] | None = None,
    checkpoint_ticks: set[int] | None = None,
) -> RunResult:
    mode = int(replay.header.game_mode_id)
    if mode == int(GameMode.SURVIVAL):
        return run_survival_replay(
            replay,
            max_ticks=max_ticks,
            warn_on_version_mismatch=warn_on_version_mismatch,
            strict_events=bool(strict_events),
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=bool(checkpoint_use_world_step_creature_count),
            checkpoints_out=checkpoints_out,
            checkpoint_ticks=checkpoint_ticks,
        )
    if mode == int(GameMode.RUSH):
        return run_rush_replay(
            replay,
            max_ticks=max_ticks,
            warn_on_version_mismatch=warn_on_version_mismatch,
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=bool(checkpoint_use_world_step_creature_count),
            checkpoints_out=checkpoints_out,
            checkpoint_ticks=checkpoint_ticks,
        )
    if mode == int(GameMode.QUESTS):
        return run_quest_replay(
            replay,
            max_ticks=max_ticks,
            warn_on_version_mismatch=warn_on_version_mismatch,
            strict_events=bool(strict_events),
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=bool(checkpoint_use_world_step_creature_count),
            checkpoints_out=checkpoints_out,
            checkpoint_ticks=checkpoint_ticks,
        )
    raise ReplayRunnerError(f"unsupported replay game_mode_id={mode}")



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace

from ..game_modes import GameMode
from ..replay.checkpoints import ReplayCheckpoint
from ..sim.driver.replay_runner import run_quest_replay, run_rush_replay, run_survival_replay
from ..sim.driver.setup import RunResult
from .capture import (
    CaptureFile,
    build_capture_dt_frame_ms_i32_overrides,
    build_capture_dt_frame_overrides,
    build_capture_inter_tick_rand_draws_overrides,
    convert_capture_to_checkpoints,
    convert_capture_to_replay,
)
from .diff import ReplayFieldDiff, checkpoint_field_diffs


class CaptureVerifyError(ValueError):
    pass


@dataclass(frozen=True, slots=True)
class CaptureVerifyFailure:
    kind: str
    tick_index: int
    expected: ReplayCheckpoint
    actual: ReplayCheckpoint | None = None
    field_diffs: tuple[ReplayFieldDiff, ...] = ()


@dataclass(frozen=True, slots=True)
class CaptureVerifyResult:
    ok: bool
    checked_count: int
    expected_count: int
    actual_count: int
    elapsed_baseline_tick: int | None = None
    elapsed_offset_ms: int | None = None
    failure: CaptureVerifyFailure | None = None


def _allow_capture_sample_creature_count(
    *,
    tick: int,
    field_diffs: list[ReplayFieldDiff],
    expected_by_tick: dict[int, ReplayCheckpoint],
    actual_by_tick: dict[int, ReplayCheckpoint],
    capture_sample_creature_counts: dict[int, int],
    capture_active_corpse_below_despawn_ticks: set[int] | None = None,
) -> bool:
    if not field_diffs:
        return False
    if any(str(diff.field) != "creature_count" for diff in field_diffs):
        return False
    if not capture_sample_creature_counts:
        return False

    sample_count = capture_sample_creature_counts.get(int(tick))
    if sample_count is None or int(sample_count) < 0:
        return False

    expected_tick = expected_by_tick.get(int(tick))
    actual_tick = actual_by_tick.get(int(tick))
    if expected_tick is None or actual_tick is None:
        return False

    expected_count = int(expected_tick.creature_count)
    actual_count = int(actual_tick.creature_count)
    if expected_count < 0 or actual_count < 0:
        return False

    # `checkpoint.creature_count` can lag the sampled creature pool in captures.
    # When replay count matches sampled active entries exactly, ignore this field.
    if actual_count == int(sample_count) and expected_count != int(sample_count):
        return True

    # Some captures expose a one-tick lead/lag between checkpoint
    # `creature_active_count` and sampled creature slots.
    next_sample = capture_sample_creature_counts.get(int(tick) + 1)
    next_expected_tick = expected_by_tick.get(int(tick) + 1)
    next_actual_tick = actual_by_tick.get(int(tick) + 1)
    if (
        next_sample is not None
        and int(next_sample) >= 0
        and abs(expected_count - actual_count) == 1
        and expected_count == int(sample_count)
        and actual_count == int(next_sample)
        and next_expected_tick is not None
        and next_actual_tick is not None
        and int(next_expected_tick.creature_count) == expected_count
        and int(next_actual_tick.creature_count) == actual_count
    ):
        return True

    # Some captures sample creature slots before render-time corpse culling.
    # In those ticks, sampled active count can exceed replay by exactly one when a
    # corpse is already below the native despawn threshold (< -10.0 hitbox_size).
    if (
        expected_count == int(sample_count)
        and actual_count == int(sample_count) - 1
        and capture_active_corpse_below_despawn_ticks is not None
        and int(tick) in capture_active_corpse_below_despawn_ticks
    ):
        return True

    return False


def _allow_one_tick_kills_lag(
    *,
    tick: int,
    field_diffs: list[ReplayFieldDiff],
    expected_by_tick: dict[int, ReplayCheckpoint],
    actual_by_tick: dict[int, ReplayCheckpoint],
) -> bool:
    if not field_diffs:
        return False
    if any(str(diff.field) != "kills" for diff in field_diffs):
        return False

    expected_tick = expected_by_tick.get(int(tick))
    actual_tick = actual_by_tick.get(int(tick))
    prev_expected = expected_by_tick.get(int(tick) - 1)
    prev_actual = actual_by_tick.get(int(tick) - 1)
    next_expected = expected_by_tick.get(int(tick) + 1)
    next_actual = actual_by_tick.get(int(tick) + 1)
    if (
        expected_tick is None
        or actual_tick is None
        or prev_expected is None
        or prev_actual is None
        or next_expected is None
        or next_actual is None
    ):
        return False

    expected_kills = int(expected_tick.kills)
    actual_kills = int(actual_tick.kills)
    if expected_kills < 0 or actual_kills < 0:
        return False
    if abs(int(expected_kills) - int(actual_kills)) != 1:
        return False

    if int(prev_expected.kills) != int(prev_actual.kills):
        return False
    if int(next_expected.kills) != int(next_actual.kills):
        return False

    return True


def _capture_sample_creature_counts(capture: CaptureFile) -> dict[int, int]:
    out: dict[int, int] = {}
    for tick in capture.ticks:
        tick_index = int(tick.tick_index)
        if tick_index < 0:
            continue
        samples = tick.samples
        if samples is None:
            continue
        creatures = samples.creatures
        if not isinstance(creatures, list):
            continue
        out[int(tick_index)] = int(len(creatures))
    return out


def _capture_active_corpse_below_despawn_ticks(capture: CaptureFile) -> set[int]:
    out: set[int] = set()
    for tick in capture.ticks:
        tick_index = int(tick.tick_index)
        if tick_index < 0:
            continue
        samples = tick.samples
        if samples is None:
            continue
        creatures = samples.creatures
        if not isinstance(creatures, list):
            continue
        for creature in creatures:
            if int(creature.active) == 0:
                continue
            if float(creature.hp) > 0.0:
                continue
            if float(creature.hitbox_size) < -10.0:
                out.add(int(tick_index))
                break
    return out


def verify_capture(
    capture: CaptureFile,
    *,
    seed: int | None = None,
    max_ticks: int | None = None,
    strict_events: bool = False,
    trace_rng: bool = False,
    max_field_diffs: int = 16,
    float_abs_tol: float = 0.001,
    aim_scheme_overrides_by_player: Mapping[int, int] | None = None,
) -> tuple[CaptureVerifyResult, RunResult]:
    expected = convert_capture_to_checkpoints(capture).checkpoints
    if max_ticks is not None:
        tick_cap = max(0, int(max_ticks))
        expected = [ckpt for ckpt in expected if int(ckpt.tick_index) < int(tick_cap)]

    replay = convert_capture_to_replay(
        capture,
        seed=seed,
        aim_scheme_overrides_by_player=aim_scheme_overrides_by_player,
    )
    dt_frame_overrides = build_capture_dt_frame_overrides(
        capture,
        tick_rate=int(replay.header.tick_rate),
    )
    dt_frame_ms_i32_overrides = build_capture_dt_frame_ms_i32_overrides(capture)
    checkpoint_ticks = {int(ckpt.tick_index) for ckpt in expected}
    actual: list[ReplayCheckpoint] = []

    mode = int(replay.header.game_mode_id)
    inter_tick_rand_draws = 1
    inter_tick_rand_draws_by_tick = build_capture_inter_tick_rand_draws_overrides(capture)

    if mode == int(GameMode.SURVIVAL):
        run_result = run_survival_replay(
            replay,
            max_ticks=max_ticks,
            strict_events=bool(strict_events),
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=False,
            checkpoints_out=actual,
            checkpoint_ticks=checkpoint_ticks,
            dt_frame_overrides=dt_frame_overrides,
            dt_frame_ms_i32_overrides=dt_frame_ms_i32_overrides,
            inter_tick_rand_draws=int(inter_tick_rand_draws),
            inter_tick_rand_draws_by_tick=inter_tick_rand_draws_by_tick,
        )
    elif mode == int(GameMode.RUSH):
        run_result = run_rush_replay(
            replay,
            max_ticks=max_ticks,
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=False,
            checkpoints_out=actual,
            checkpoint_ticks=checkpoint_ticks,
            dt_frame_overrides=dt_frame_overrides,
            inter_tick_rand_draws=int(inter_tick_rand_draws),
            inter_tick_rand_draws_by_tick=inter_tick_rand_draws_by_tick,
        )
    elif mode == int(GameMode.QUESTS):
        run_result = run_quest_replay(
            replay,
            max_ticks=max_ticks,
            strict_events=bool(strict_events),
            trace_rng=bool(trace_rng),
            checkpoint_use_world_step_creature_count=False,
            checkpoints_out=actual,
            checkpoint_ticks=checkpoint_ticks,
            dt_frame_overrides=dt_frame_overrides,
            inter_tick_rand_draws=int(inter_tick_rand_draws),
            inter_tick_rand_draws_by_tick=inter_tick_rand_draws_by_tick,
        )
    else:
        raise CaptureVerifyError(f"unsupported game mode for capture verification: {mode}")

    expected_by_tick = {int(ckpt.tick_index): ckpt for ckpt in expected}
    actual_by_tick = {int(ckpt.tick_index): ckpt for ckpt in actual}
    sample_creature_counts = _capture_sample_creature_counts(capture)
    sample_corpse_below_despawn_ticks = _capture_active_corpse_below_despawn_ticks(capture)
    checked_count = 0
    elapsed_baseline: tuple[int, int] | None = None
    elapsed_baseline_tick: int | None = None
    elapsed_offset_ms: int | None = None

    for exp in expected:
        checked_count += 1
        tick = int(exp.tick_index)
        act = actual_by_tick.get(tick)
        if act is None:
            return (
                CaptureVerifyResult(
                    ok=False,
                    checked_count=checked_count,
                    expected_count=len(expected),
                    actual_count=len(actual),
                    elapsed_baseline_tick=elapsed_baseline_tick,
                    elapsed_offset_ms=elapsed_offset_ms,
                    failure=CaptureVerifyFailure(
                        kind="missing_checkpoint",
                        tick_index=tick,
                        expected=exp,
                        actual=None,
                    ),
                ),
                run_result,
            )

        if elapsed_baseline is None and int(exp.elapsed_ms) >= 0 and int(act.elapsed_ms) >= 0:
            elapsed_baseline = (int(exp.elapsed_ms), int(act.elapsed_ms))
            elapsed_baseline_tick = int(tick)
            elapsed_offset_ms = int(act.elapsed_ms) - int(exp.elapsed_ms)

        exp_for_diff = replace(exp, elapsed_ms=-1)
        field_diffs = checkpoint_field_diffs(
            exp_for_diff,
            act,
            include_hash_fields=False,
            include_rng_fields=False,
            normalize_unknown=True,
            unknown_events_wildcard=True,
            elapsed_baseline=elapsed_baseline,
            max_diffs=max_field_diffs,
            float_abs_tol=float(float_abs_tol),
        )
        if _allow_capture_sample_creature_count(
            tick=int(tick),
            field_diffs=field_diffs,
            expected_by_tick=expected_by_tick,
            actual_by_tick=actual_by_tick,
            capture_sample_creature_counts=sample_creature_counts,
            capture_active_corpse_below_despawn_ticks=sample_corpse_below_despawn_ticks,
        ):
            continue
        if _allow_one_tick_kills_lag(
            tick=int(tick),
            field_diffs=field_diffs,
            expected_by_tick=expected_by_tick,
            actual_by_tick=actual_by_tick,
        ):
            continue
        if field_diffs:
            return (
                CaptureVerifyResult(
                    ok=False,
                    checked_count=checked_count,
                    expected_count=len(expected),
                    actual_count=len(actual),
                    elapsed_baseline_tick=elapsed_baseline_tick,
                    elapsed_offset_ms=elapsed_offset_ms,
                    failure=CaptureVerifyFailure(
                        kind="state_mismatch",
                        tick_index=tick,
                        expected=exp,
                        actual=act,
                        field_diffs=tuple(field_diffs),
                    ),
                ),
                run_result,
            )

    return (
        CaptureVerifyResult(
            ok=True,
            checked_count=checked_count,
            expected_count=len(expected),
            actual_count=len(actual),
            elapsed_baseline_tick=elapsed_baseline_tick,
            elapsed_offset_ms=elapsed_offset_ms,
            failure=None,
        ),
        run_result,
    )



from __future__ import annotations

from .primary_beam import draw_beam_effect
from .primary_bullet import draw_bullet_trail
from .primary_plasma import draw_plasma_particles
from .primary_special import draw_plague_spreader, draw_pulse_gun, draw_splitter_or_blade
from .types import ProjectileDrawCtx

PRIMARY_PROJECTILE_DRAW_HANDLERS = (
    draw_bullet_trail,
    draw_plasma_particles,
    draw_beam_effect,
    draw_pulse_gun,
    draw_splitter_or_blade,
    draw_plague_spreader,
)


def draw_projectile_from_registry(ctx: ProjectileDrawCtx) -> bool:
    for handler in PRIMARY_PROJECTILE_DRAW_HANDLERS:
        if handler(ctx):
            return True
    return False


__all__ = ["draw_projectile_from_registry"]



from __future__ import annotations

from ..math_parity import f32
from .apply_context import BonusApplyCtx, bonus_apply_seconds


def apply_energizer(ctx: BonusApplyCtx) -> None:
    old = float(ctx.state.bonuses.energizer)
    if old <= 0.0:
        ctx.register_global("energizer")

    ctx.state.bonuses.energizer = float(
        f32(float(old) + bonus_apply_seconds(ctx) * float(ctx.economist_multiplier)),
    )



from __future__ import annotations

import pyray as rl
from raylib import defines as rd

from grim.color import RGBA
from grim.math import clamp

from ...effects_atlas import EFFECT_ID_ATLAS_TABLE_BY_ID, SIZE_CODE_GRID, EffectId
from ...perks import PerkId
from ...perks.helpers import perk_active
from ...projectiles import ProjectileTypeId
from ...sim.world_defs import BEAM_TYPES, ION_TYPES
from ..projectile_render_registry import beam_effect_scale
from .common import RAD_TO_DEG, proj_origin
from .types import ProjectileDrawCtx


def draw_beam_effect(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    type_id = int(ctx.type_id)
    texture = ctx.texture
    if type_id not in BEAM_TYPES:
        return False
    if texture is None:
        return False

    # Ion weapons and Fire Bullets use the projs.png streak effect (and Ion adds chain arcs on impact).
    grid = 4
    frame = 2

    is_fire_bullets = type_id == int(ProjectileTypeId.FIRE_BULLETS)
    is_ion = type_id in ION_TYPES

    origin = proj_origin(ctx.proj, ctx.pos)
    beam = ctx.pos - origin
    direction, dist = beam.normalized_with_length()
    if dist <= 1e-6:
        return True

    # In the native renderer, Ion Gun Master increases the chain effect thickness and reach.
    perk_scale = 1.0
    if any(perk_active(player, PerkId.ION_GUN_MASTER) for player in renderer.players):
        perk_scale = 1.2

    effect_scale = beam_effect_scale(type_id)

    alpha = float(ctx.alpha)
    life = float(ctx.life)
    if life >= 0.4:
        base_alpha = alpha
    else:
        fade = clamp(life * 2.5, 0.0, 1.0)
        base_alpha = fade * alpha

    if base_alpha <= 1e-3:
        return True

    streak_rgb = (1.0, 0.6, 0.1) if is_fire_bullets else (0.5, 0.6, 1.0)
    head_rgb = (1.0, 1.0, 0.7)

    # Only draw the last 256 units of the path.
    start = 0.0
    span = dist
    if dist > 256.0:
        start = dist - 256.0
        span = 256.0

    step = min(effect_scale * 3.1, 9.0)
    sprite_scale = effect_scale * ctx.scale

    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)

    s = start
    while s < dist:
        t = (s - start) / span if span > 1e-6 else 1.0
        seg_alpha = t * base_alpha
        if seg_alpha > 1e-3:
            pos = origin + direction * s
            pos_screen = renderer.world_to_screen(pos)
            tint = RGBA(streak_rgb[0], streak_rgb[1], streak_rgb[2], seg_alpha).to_rl()
            renderer._draw_atlas_sprite(
                texture,
                grid=grid,
                frame=frame,
                pos=pos_screen,
                scale=sprite_scale,
                rotation_rad=ctx.angle,
                tint=tint,
            )
        s += step

    if life >= 0.4:
        head_tint = RGBA(head_rgb[0], head_rgb[1], head_rgb[2], base_alpha).to_rl()
        renderer._draw_atlas_sprite(
            texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=sprite_scale,
            rotation_rad=ctx.angle,
            tint=head_tint,
        )

        # Fire Bullets renders an extra particles.png overlay in a later pass.
        if is_fire_bullets and renderer.particles_texture is not None:
            particles_texture = renderer.particles_texture
            atlas = EFFECT_ID_ATLAS_TABLE_BY_ID.get(int(EffectId.GLOW))
            if atlas is not None:
                grid = SIZE_CODE_GRID.get(int(atlas.size_code))
                if grid:
                    cell_w = float(particles_texture.width) / float(grid)
                    cell_h = float(particles_texture.height) / float(grid)
                    frame = int(atlas.frame)
                    col = frame % grid
                    row = frame // grid
                    src = rl.Rectangle(
                        cell_w * float(col),
                        cell_h * float(row),
                        max(0.0, cell_w - 2.0),
                        max(0.0, cell_h - 2.0),
                    )
                    tint = RGBA(1.0, 1.0, 1.0, alpha).to_rl()
                    size = 64.0 * ctx.scale
                    dst = rl.Rectangle(ctx.screen_pos.x, ctx.screen_pos.y, float(size), float(size))
                    origin = rl.Vector2(size * 0.5, size * 0.5)
                    rl.draw_texture_pro(particles_texture, src, dst, origin, ctx.angle * RAD_TO_DEG, tint)
    else:
        # Native draws a small blue "core" at the head during the fade stage (life_timer < 0.4).
        core_tint = RGBA(0.5, 0.6, 1.0, base_alpha).to_rl()
        renderer._draw_atlas_sprite(
            texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=1.0 * ctx.scale,
            rotation_rad=ctx.angle,
            tint=core_tint,
        )

        if is_ion:
            # Native: chain reach is derived from the streak scale (`fVar29 * perk_scale * 40.0`).
            radius = effect_scale * perk_scale * 40.0

            # Native iterates via creature_find_in_radius(pos, radius, start_index) in pool order.
            targets = []
            for creature in renderer.creatures.entries[1:]:
                if not creature.active:
                    continue
                if float(getattr(creature, "hitbox_size", 0.0)) <= 5.0:
                    continue
                d = ctx.pos.distance_to(creature.pos)
                threshold = float(creature.size) * 0.14285715 + 3.0
                if d - radius < threshold:
                    targets.append(creature)

            inner_half = 10.0 * perk_scale * ctx.scale
            outer_half = 14.0 * perk_scale * ctx.scale
            u = 0.625
            v0 = 0.0
            v1 = 0.25

            glow_targets = []
            rl.rl_set_texture(texture.id)
            rl.rl_begin(rd.RL_QUADS)

            for creature in targets:
                target_screen = renderer.world_to_screen(creature.pos)
                segment = target_screen - ctx.screen_pos
                direction_screen, dlen = segment.normalized_with_length()
                if dlen <= 1e-3:
                    continue
                glow_targets.append(creature)
                side = direction_screen.perp_left()

                # Outer strip (softer).
                side_offset = side * outer_half
                p0 = ctx.screen_pos - side_offset
                p1 = ctx.screen_pos + side_offset
                p2 = target_screen + side_offset
                p3 = target_screen - side_offset

                outer_tint = RGBA(0.5, 0.6, 1.0, base_alpha).to_rl()
                rl.rl_color4ub(outer_tint.r, outer_tint.g, outer_tint.b, outer_tint.a)
                rl.rl_tex_coord2f(u, v0)
                rl.rl_vertex2f(p0.x, p0.y)
                rl.rl_tex_coord2f(u, v1)
                rl.rl_vertex2f(p1.x, p1.y)
                rl.rl_tex_coord2f(u, v1)
                rl.rl_vertex2f(p2.x, p2.y)
                rl.rl_tex_coord2f(u, v0)
                rl.rl_vertex2f(p3.x, p3.y)

                # Inner strip (brighter).
                side_offset = side * inner_half
                p0 = ctx.screen_pos - side_offset
                p1 = ctx.screen_pos + side_offset
                p2 = target_screen + side_offset
                p3 = target_screen - side_offset

                inner_tint = RGBA(0.5, 0.6, 1.0, base_alpha).to_rl()
                rl.rl_color4ub(inner_tint.r, inner_tint.g, inner_tint.b, inner_tint.a)
                rl.rl_tex_coord2f(u, v0)
                rl.rl_vertex2f(p0.x, p0.y)
                rl.rl_tex_coord2f(u, v1)
                rl.rl_vertex2f(p1.x, p1.y)
                rl.rl_tex_coord2f(u, v1)
                rl.rl_vertex2f(p2.x, p2.y)
                rl.rl_tex_coord2f(u, v0)
                rl.rl_vertex2f(p3.x, p3.y)

            rl.rl_end()
            rl.rl_set_texture(0)

            for creature in glow_targets:
                target_screen = renderer.world_to_screen(creature.pos)
                target_tint = RGBA(0.5, 0.6, 1.0, base_alpha).to_rl()
                renderer._draw_atlas_sprite(
                    texture,
                    grid=grid,
                    frame=frame,
                    pos=target_screen,
                    scale=sprite_scale,
                    rotation_rad=0.0,
                    tint=target_tint,
                )

    rl.end_blend_mode()
    return True


__all__ = ["draw_beam_effect"]



from __future__ import annotations

import random
import string
import time
import uuid
from dataclasses import dataclass, field

import structlog

from ..logging import ensure_structlog_stdlib_defaults
from .relay_protocol import (
    LINK_TIMEOUT_MS,
    PROTOCOL_VERSION,
    RECONNECT_TIMEOUT_MS,
    ROOM_CODE_LENGTH,
    ClientHello,
    ClientWelcome,
    LegacyLockstepControl,
    LegacyLockstepInputBatch,
    LegacyLockstepTickFrame,
    NetcodeMode,
    NetMessage,
    PeerDisconnect,
    Ping,
    Pong,
    RbInputBatch,
    RbResyncBegin,
    RbResyncChunk,
    RbResyncCommit,
    RbResyncRequest,
    RelayError,
    RelaySlot,
    RoomCreate,
    RoomJoin,
    RoomReady,
    RoomStart,
    RoomState,
    StatusSnapshot,
    builds_compatible,
)
from .relay_reliable import RelayReliableLink
from .relay_transport import PeerAddr, RelayUdpTransport


def _now_ms() -> int:
    return int(time.monotonic() * 1000.0)


def _room_code(rand: random.Random) -> str:
    alphabet = string.ascii_uppercase + string.digits
    return "".join(rand.choice(alphabet) for _ in range(int(ROOM_CODE_LENGTH)))


@dataclass(slots=True)
class RelayServerConfig:
    bind_host: str = "0.0.0.0"
    bind_port: int = 31993
    link_timeout_ms: int = LINK_TIMEOUT_MS
    reconnect_timeout_ms: int = RECONNECT_TIMEOUT_MS
    max_rooms: int = 2048


@dataclass(slots=True)
class _Peer:
    addr: PeerAddr
    link: RelayReliableLink = field(default_factory=RelayReliableLink)
    peer_id: str = ""
    build_id: str = ""
    peer_name: str = ""
    room_code: str = ""
    slot_index: int = -1
    last_seen_ms: int = 0


@dataclass(slots=True)
class _RoomSlot:
    slot_index: int
    peer_id: str = ""
    peer_name: str = ""
    ready: bool = False
    reconnect_token: str = ""
    disconnected_ms: int = 0

    @property
    def connected(self) -> bool:
        return bool(self.peer_id)


@dataclass(slots=True)
class _Room:
    room_code: str
    session_id: str
    mode_id: int
    player_count: int
    quest_level: str
    preserve_bugs: bool
    tick_rate: int
    input_delay_ticks: int
    rollback_max_ticks: int
    netcode_mode: NetcodeMode
    status_snapshot: StatusSnapshot | None = None
    started: bool = False
    seed: int = 0
    start_tick: int = 0
    slots: list[_RoomSlot] = field(default_factory=list)

    def all_ready(self) -> bool:
        for slot in self.slots:
            if not slot.connected:
                return False
            if not bool(slot.ready):
                return False
        return True


_FORWARD_RELIABLE_TYPES = (
    RbResyncRequest,
    RbResyncBegin,
    RbResyncChunk,
    RbResyncCommit,
    LegacyLockstepTickFrame,
    LegacyLockstepControl,
)


class RelayServer:
    """In-memory UDP relay + invite-code lobby service."""

    def __init__(self, cfg: RelayServerConfig) -> None:
        ensure_structlog_stdlib_defaults()
        self.cfg = cfg
        self.transport = RelayUdpTransport(bind_host=str(cfg.bind_host), bind_port=int(cfg.bind_port))
        self._rooms: dict[str, _Room] = {}
        self._peers_by_addr: dict[PeerAddr, _Peer] = {}
        self._peers_by_id: dict[str, _Peer] = {}
        self._room_by_reconnect: dict[str, tuple[str, int]] = {}
        self._rand = random.Random()
        self.log = structlog.stdlib.get_logger("crimson.relay").bind(
            component="relay_service",
            bind_host=str(cfg.bind_host),
            bind_port=int(cfg.bind_port),
        )

    @property
    def bound_port(self) -> int:
        return int(self.transport.bound_port)

    def open(self) -> None:
        self.transport.open()
        self.log.info(
            "relay_open",
            bound_port=int(self.bound_port),
            link_timeout_ms=int(self.cfg.link_timeout_ms),
            reconnect_timeout_ms=int(self.cfg.reconnect_timeout_ms),
            max_rooms=int(self.cfg.max_rooms),
        )

    def close(self) -> None:
        self.transport.close()
        room_count = len(self._rooms)
        peer_count = len(self._peers_by_id)
        self._rooms.clear()
        self._peers_by_addr.clear()
        self._peers_by_id.clear()
        self._room_by_reconnect.clear()
        self.log.info("relay_close", room_count=int(room_count), peer_count=int(peer_count))

    def serve_forever(self, *, tick_ms: int = 8) -> None:
        self.open()
        try:
            while True:
                now = _now_ms()
                self.update(now_ms=int(now))
                time.sleep(max(0.0, float(tick_ms) * 0.001))
        finally:
            self.close()

    def update(self, *, now_ms: int | None = None) -> None:
        if now_ms is None:
            now_ms = _now_ms()
        now = int(now_ms)

        for addr, packet in self.transport.recv_packets(max_packets=512):
            peer = self._peers_by_addr.get(addr)
            if peer is None:
                if not isinstance(packet.message, ClientHello):
                    self.log.debug(
                        "relay_drop_unregistered_packet",
                        addr=self._addr_text(addr),
                        kind=type(packet.message).__name__,
                        reliable=bool(packet.reliable),
                        seq=int(packet.seq),
                    )
                    continue
                self._handle_client_hello(addr=addr, message=packet.message, now_ms=int(now))
                peer = self._peers_by_addr.get(addr)
                if peer is not None and bool(packet.reliable) and int(packet.seq) > 0:
                    peer.link.prime_recv_seq(int(packet.seq))
                    self.log.debug(
                        "relay_prime_recv_seq",
                        peer_id=str(peer.peer_id),
                        seq=int(packet.seq),
                    )
                continue

            peer.last_seen_ms = int(now)
            messages, duplicate = peer.link.ingest_packet(packet, now_ms=int(now))
            if bool(duplicate):
                self.log.debug(
                    "relay_recv_duplicate",
                    peer_id=str(peer.peer_id),
                    addr=self._addr_text(addr),
                    seq=int(packet.seq),
                    ack=int(packet.ack),
                )
            for message in messages:
                self._handle_message(peer=peer, message=message, now_ms=int(now))

        self._poll_resends(now_ms=int(now))
        self._prune_timeouts(now_ms=int(now))

    def _handle_client_hello(self, *, addr: PeerAddr, message: ClientHello, now_ms: int) -> None:
        if int(message.protocol_version) != int(PROTOCOL_VERSION):
            self.log.warning(
                "relay_protocol_mismatch",
                addr=self._addr_text(addr),
                peer_protocol=int(message.protocol_version),
                expected_protocol=int(PROTOCOL_VERSION),
                build_id=str(message.build_id or ""),
            )
            temp = RelayReliableLink()
            packet = temp.build_packet(
                ClientWelcome(
                    accepted=False,
                    reason="protocol_mismatch_v5_required",
                    protocol_version=int(PROTOCOL_VERSION),
                ),
                reliable=True,
                now_ms=int(now_ms),
            )
            self.transport.send_packet(addr, packet)
            return

        peer_id = uuid.uuid4().hex[:12]
        peer = _Peer(
            addr=addr,
            peer_id=str(peer_id),
            build_id=str(message.build_id or ""),
            peer_name=str(message.peer_name or ""),
            last_seen_ms=int(now_ms),
        )
        self._peers_by_addr[addr] = peer
        self._peers_by_id[str(peer_id)] = peer
        self.log.info(
            "relay_peer_connected",
            peer_id=str(peer_id),
            addr=self._addr_text(addr),
            build_id=str(message.build_id or ""),
            peer_name=str(message.peer_name or ""),
            peer_count=int(len(self._peers_by_id)),
        )

        self._send_peer(
            peer,
            ClientWelcome(
                accepted=True,
                reason="",
                protocol_version=int(PROTOCOL_VERSION),
                build_id=str(message.build_id or ""),
                peer_id=str(peer_id),
            ),
            reliable=True,
            now_ms=int(now_ms),
        )

    def _handle_message(self, *, peer: _Peer, message: NetMessage, now_ms: int) -> None:
        self.log.debug(
            "relay_message_received",
            peer_id=str(peer.peer_id),
            room_code=str(peer.room_code or ""),
            slot_index=int(peer.slot_index),
            kind=type(message).__name__,
            request_id=str(getattr(message, "request_id", "") or ""),
            now_ms=int(now_ms),
        )
        if isinstance(message, Ping):
            self._send_peer(peer, Pong(stamp_ms=int(message.stamp_ms)), reliable=False, now_ms=int(now_ms))
            return

        if isinstance(message, RoomCreate):
            self._handle_room_create(peer=peer, message=message, now_ms=int(now_ms))
            return

        if isinstance(message, RoomJoin):
            self._handle_room_join(peer=peer, message=message, now_ms=int(now_ms))
            return

        if isinstance(message, RoomReady):
            self._handle_room_ready(peer=peer, message=message, now_ms=int(now_ms))
            return

        if isinstance(message, (RbInputBatch, RbResyncRequest, RbResyncBegin, RbResyncChunk, RbResyncCommit)):
            self._forward_room_message(peer=peer, message=message, now_ms=int(now_ms))
            return

        if isinstance(message, (LegacyLockstepInputBatch, LegacyLockstepTickFrame, LegacyLockstepControl)):
            self._forward_room_message(peer=peer, message=message, now_ms=int(now_ms))
            return

        self.log.warning(
            "relay_message_unhandled",
            peer_id=str(peer.peer_id),
            kind=type(message).__name__,
            room_code=str(peer.room_code or ""),
        )

    def _handle_room_create(self, *, peer: _Peer, message: RoomCreate, now_ms: int) -> None:
        if peer.room_code:
            self.log.warning(
                "relay_room_create_rejected",
                reason="already_in_room",
                peer_id=str(peer.peer_id),
                room_code=str(peer.room_code or ""),
            )
            self._send_peer(peer, RelayError(reason="already_in_room"), reliable=True, now_ms=int(now_ms))
            return
        if len(self._rooms) >= int(self.cfg.max_rooms):
            self.log.warning(
                "relay_room_create_rejected",
                reason="room_capacity",
                peer_id=str(peer.peer_id),
                room_count=int(len(self._rooms)),
                max_rooms=int(self.cfg.max_rooms),
            )
            self._send_peer(peer, RelayError(reason="room_capacity"), reliable=True, now_ms=int(now_ms))
            return

        code = ""
        for _ in range(32):
            candidate = _room_code(self._rand)
            if candidate in self._rooms:
                continue
            code = candidate
            break
        if not code:
            self.log.error(
                "relay_room_create_rejected",
                reason="room_code_exhausted",
                peer_id=str(peer.peer_id),
            )
            self._send_peer(peer, RelayError(reason="room_code_exhausted"), reliable=True, now_ms=int(now_ms))
            return

        player_count = max(1, min(4, int(message.player_count)))
        slots = [_RoomSlot(slot_index=i) for i in range(int(player_count))]
        host_slot = slots[0]
        host_slot.peer_id = str(peer.peer_id)
        host_slot.peer_name = str(peer.peer_name)
        host_slot.ready = True
        host_slot.reconnect_token = uuid.uuid4().hex

        room = _Room(
            room_code=str(code),
            session_id=uuid.uuid4().hex[:12],
            mode_id=int(message.mode_id),
            player_count=int(player_count),
            quest_level=str(message.quest_level or ""),
            preserve_bugs=bool(message.preserve_bugs),
            tick_rate=max(1, int(message.tick_rate)),
            input_delay_ticks=max(0, int(message.input_delay_ticks)),
            rollback_max_ticks=max(1, int(message.rollback_max_ticks)),
            netcode_mode=message.netcode_mode,
            status_snapshot=message.status_snapshot,
            slots=slots,
        )
        self._rooms[str(code)] = room
        self._room_by_reconnect[str(host_slot.reconnect_token)] = (str(code), 0)

        peer.room_code = str(code)
        peer.slot_index = 0
        self.log.info(
            "relay_room_created",
            room_code=str(room.room_code),
            session_id=str(room.session_id),
            host_peer_id=str(peer.peer_id),
            mode_id=int(room.mode_id),
            player_count=int(room.player_count),
            netcode_mode=str(room.netcode_mode),
        )

        self._broadcast_room_state(room=room, now_ms=int(now_ms))

    def _handle_room_join(self, *, peer: _Peer, message: RoomJoin, now_ms: int) -> None:
        room_code = str(message.room_code or "").upper().strip()
        reconnect_token = str(message.reconnect_token or "").strip()

        if peer.room_code:
            self.log.warning(
                "relay_room_join_rejected",
                reason="already_in_room",
                peer_id=str(peer.peer_id),
                room_code=str(peer.room_code or ""),
            )
            self._send_peer(peer, RelayError(reason="already_in_room"), reliable=True, now_ms=int(now_ms))
            return

        room: _Room | None = None
        slot_index = -1

        if reconnect_token:
            mapping = self._room_by_reconnect.get(str(reconnect_token))
            if mapping is not None:
                room = self._rooms.get(str(mapping[0]))
                slot_index = int(mapping[1])

        if room is None:
            room = self._rooms.get(str(room_code))
            if room is None:
                self.log.warning(
                    "relay_room_join_rejected",
                    reason="room_not_found",
                    peer_id=str(peer.peer_id),
                    room_code=str(room_code),
                )
                self._send_peer(peer, RelayError(reason="room_not_found"), reliable=True, now_ms=int(now_ms))
                return
            slot_index = self._next_free_slot(room)
            if slot_index < 0:
                self.log.warning(
                    "relay_room_join_rejected",
                    reason="room_full",
                    peer_id=str(peer.peer_id),
                    room_code=str(room.room_code),
                )
                self._send_peer(peer, RelayError(reason="room_full"), reliable=True, now_ms=int(now_ms))
                return

        slot = room.slots[int(slot_index)]
        if slot.peer_id and slot.peer_id != str(peer.peer_id):
            age = max(0, int(now_ms) - int(slot.disconnected_ms or 0))
            if age < int(self.cfg.reconnect_timeout_ms):
                self.log.warning(
                    "relay_room_join_rejected",
                    reason="slot_busy",
                    peer_id=str(peer.peer_id),
                    room_code=str(room.room_code),
                    slot_index=int(slot_index),
                    age_ms=int(age),
                )
                self._send_peer(peer, RelayError(reason="slot_busy"), reliable=True, now_ms=int(now_ms))
                return

        host_slot = room.slots[0]
        host_peer = self._peers_by_id.get(str(host_slot.peer_id))
        if host_peer is not None and host_peer.build_id and peer.build_id:
            if not builds_compatible(str(peer.build_id), str(host_peer.build_id)):
                self.log.warning(
                    "relay_room_join_rejected",
                    reason="build_mismatch",
                    peer_id=str(peer.peer_id),
                    room_code=str(room.room_code),
                    peer_build=str(peer.build_id),
                    host_build=str(host_peer.build_id),
                )
                self._send_peer(peer, RelayError(reason="build_mismatch"), reliable=True, now_ms=int(now_ms))
                return

        slot.peer_id = str(peer.peer_id)
        slot.peer_name = str(peer.peer_name)
        slot.ready = False if not room.started else True
        slot.disconnected_ms = 0
        if not slot.reconnect_token:
            slot.reconnect_token = uuid.uuid4().hex
        self._room_by_reconnect[str(slot.reconnect_token)] = (str(room.room_code), int(slot.slot_index))

        peer.room_code = str(room.room_code)
        peer.slot_index = int(slot.slot_index)
        self.log.info(
            "relay_room_joined",
            room_code=str(room.room_code),
            peer_id=str(peer.peer_id),
            slot_index=int(slot.slot_index),
            reconnect=bool(reconnect_token),
            room_started=bool(room.started),
        )

        self._broadcast_room_state(room=room, now_ms=int(now_ms))

        if room.started:
            self._send_room_start(peer=peer, room=room, now_ms=int(now_ms))

    def _handle_room_ready(self, *, peer: _Peer, message: RoomReady, now_ms: int) -> None:
        room = self._rooms.get(str(peer.room_code))
        if room is None:
            self.log.warning(
                "relay_room_ready_rejected",
                reason="not_in_room",
                peer_id=str(peer.peer_id),
            )
            self._send_peer(peer, RelayError(reason="not_in_room"), reliable=True, now_ms=int(now_ms))
            return
        slot_index = int(peer.slot_index)
        if slot_index < 0 or slot_index >= len(room.slots):
            self.log.warning(
                "relay_room_ready_rejected",
                reason="bad_slot",
                peer_id=str(peer.peer_id),
                slot_index=int(slot_index),
                room_code=str(room.room_code),
            )
            self._send_peer(peer, RelayError(reason="bad_slot"), reliable=True, now_ms=int(now_ms))
            return
        slot = room.slots[int(slot_index)]
        if slot.peer_id != str(peer.peer_id):
            self.log.warning(
                "relay_room_ready_rejected",
                reason="slot_owner_mismatch",
                peer_id=str(peer.peer_id),
                slot_index=int(slot_index),
                room_code=str(room.room_code),
            )
            self._send_peer(peer, RelayError(reason="slot_owner_mismatch"), reliable=True, now_ms=int(now_ms))
            return
        slot.ready = bool(message.ready)
        self.log.debug(
            "relay_room_ready_updated",
            room_code=str(room.room_code),
            peer_id=str(peer.peer_id),
            slot_index=int(slot_index),
            ready=bool(slot.ready),
        )

        if (not room.started) and room.all_ready():
            room.started = True
            room.seed = int((now_ms * 1103515245 + 12345) & 0xFFFFFFFF)
            room.start_tick = 0
            self.log.info(
                "relay_room_started",
                room_code=str(room.room_code),
                session_id=str(room.session_id),
                seed=int(room.seed),
                player_count=int(room.player_count),
            )
            self._broadcast_room_start(room=room, now_ms=int(now_ms))

        self._broadcast_room_state(room=room, now_ms=int(now_ms))

    def _forward_room_message(self, *, peer: _Peer, message: NetMessage, now_ms: int) -> None:
        room = self._rooms.get(str(peer.room_code))
        if room is None:
            self.log.warning(
                "relay_forward_rejected",
                reason="not_in_room",
                peer_id=str(peer.peer_id),
                kind=type(message).__name__,
            )
            self._send_peer(peer, RelayError(reason="not_in_room"), reliable=True, now_ms=int(now_ms))
            return

        sender_slot = int(peer.slot_index)
        host_slot = 0
        if int(sender_slot) < 0 or int(sender_slot) >= len(room.slots):
            self.log.warning(
                "relay_forward_rejected",
                reason="bad_sender_slot",
                room_code=str(room.room_code),
                from_peer_id=str(peer.peer_id),
                kind=type(message).__name__,
                sender_slot=int(sender_slot),
            )
            self._send_peer(peer, RelayError(reason="bad_sender_slot"), reliable=True, now_ms=int(now_ms))
            return

        request_id = ""
        if isinstance(message, RbResyncRequest):
            request_id = str(message.request_id or "")
            if int(sender_slot) == int(host_slot):
                self.log.warning(
                    "relay_forward_rejected",
                    reason="invalid_resync_sender",
                    room_code=str(room.room_code),
                    from_peer_id=str(peer.peer_id),
                    kind=type(message).__name__,
                    sender_slot=int(sender_slot),
                    request_id=str(request_id),
                )
                self._send_peer(peer, RelayError(reason="invalid_resync_sender"), reliable=True, now_ms=int(now_ms))
                return
        elif isinstance(message, (RbResyncBegin, RbResyncChunk, RbResyncCommit)):
            request_id = str(getattr(message, "request_id", "") or "")
            if int(sender_slot) != int(host_slot):
                self.log.warning(
                    "relay_forward_rejected",
                    reason="invalid_resync_sender",
                    room_code=str(room.room_code),
                    from_peer_id=str(peer.peer_id),
                    kind=type(message).__name__,
                    sender_slot=int(sender_slot),
                    request_id=str(request_id),
                )
                self._send_peer(peer, RelayError(reason="invalid_resync_sender"), reliable=True, now_ms=int(now_ms))
                return

        reliable = isinstance(message, _FORWARD_RELIABLE_TYPES)
        if isinstance(message, LegacyLockstepInputBatch):
            reliable = False
        forwarded = 0
        for slot in room.slots:
            if (not slot.connected) or slot.peer_id == str(peer.peer_id):
                continue
            dst = self._peers_by_id.get(str(slot.peer_id))
            if dst is None:
                continue
            self._send_peer(dst, message, reliable=bool(reliable), now_ms=int(now_ms))
            forwarded += 1
        self.log.debug(
            "relay_forwarded",
            room_code=str(room.room_code),
            from_peer_id=str(peer.peer_id),
            kind=type(message).__name__,
            reliable=bool(reliable),
            recipient_count=int(forwarded),
            request_id=str(request_id),
        )

    def _broadcast_room_state(self, *, room: _Room, now_ms: int) -> None:
        slots = [
            RelaySlot(
                slot_index=int(slot.slot_index),
                connected=bool(slot.connected),
                ready=bool(slot.ready),
                is_host=int(slot.slot_index) == 0,
                peer_name=str(slot.peer_name or ""),
            )
            for slot in room.slots
        ]
        state = RoomState(
            room_code=str(room.room_code),
            session_id=str(room.session_id),
            mode_id=int(room.mode_id),
            player_count=int(room.player_count),
            quest_level=str(room.quest_level),
            preserve_bugs=bool(room.preserve_bugs),
            tick_rate=int(room.tick_rate),
            input_delay_ticks=int(room.input_delay_ticks),
            rollback_max_ticks=int(room.rollback_max_ticks),
            netcode_mode=room.netcode_mode,
            slots=slots,
            all_ready=bool(room.all_ready()),
            started=bool(room.started),
        )
        connected = 0
        for slot in room.slots:
            if not slot.connected:
                continue
            peer = self._peers_by_id.get(str(slot.peer_id))
            if peer is None:
                continue
            connected += 1
            self._send_peer(peer, state, reliable=True, now_ms=int(now_ms))
        self.log.debug(
            "relay_room_state_broadcast",
            room_code=str(room.room_code),
            started=bool(room.started),
            all_ready=bool(room.all_ready()),
            connected_players=int(connected),
            expected_players=int(room.player_count),
        )

    def _broadcast_room_start(self, *, room: _Room, now_ms: int) -> None:
        for slot in room.slots:
            if not slot.connected:
                continue
            peer = self._peers_by_id.get(str(slot.peer_id))
            if peer is None:
                continue
            self._send_room_start(peer=peer, room=room, now_ms=int(now_ms))

    def _send_room_start(self, *, peer: _Peer, room: _Room, now_ms: int) -> None:
        slot_index = int(peer.slot_index)
        reconnect_token = ""
        if 0 <= int(slot_index) < len(room.slots):
            reconnect_token = str(room.slots[int(slot_index)].reconnect_token)
        self._send_peer(
            peer,
            RoomStart(
                room_code=str(room.room_code),
                session_id=str(room.session_id),
                seed=int(room.seed),
                start_tick=int(room.start_tick),
                mode_id=int(room.mode_id),
                player_count=int(room.player_count),
                quest_level=str(room.quest_level),
                preserve_bugs=bool(room.preserve_bugs),
                tick_rate=int(room.tick_rate),
                input_delay_ticks=int(room.input_delay_ticks),
                rollback_max_ticks=int(room.rollback_max_ticks),
                netcode_mode=room.netcode_mode,
                slot_index=int(slot_index),
                host_slot_index=0,
                reconnect_token=str(reconnect_token),
                status_snapshot=room.status_snapshot,
            ),
            reliable=True,
            now_ms=int(now_ms),
        )
        self.log.info(
            "relay_room_start_sent",
            room_code=str(room.room_code),
            session_id=str(room.session_id),
            peer_id=str(peer.peer_id),
            slot_index=int(slot_index),
        )

    def _send_peer(self, peer: _Peer, message: NetMessage, *, reliable: bool, now_ms: int) -> None:
        packet = peer.link.build_packet(message, reliable=bool(reliable), now_ms=int(now_ms))
        self.transport.send_packet(peer.addr, packet)
        self.log.debug(
            "relay_send",
            peer_id=str(peer.peer_id),
            addr=self._addr_text(peer.addr),
            room_code=str(peer.room_code or ""),
            slot_index=int(peer.slot_index),
            kind=type(message).__name__,
            reliable=bool(reliable),
            seq=int(packet.seq),
            ack=int(packet.ack),
        )

    def _poll_resends(self, *, now_ms: int) -> None:
        for peer in self._peers_by_id.values():
            for packet in peer.link.poll_resends(now_ms=int(now_ms)):
                self.transport.send_packet(peer.addr, packet)
                self.log.debug(
                    "relay_resend",
                    peer_id=str(peer.peer_id),
                    addr=self._addr_text(peer.addr),
                    kind=type(packet.message).__name__,
                    seq=int(packet.seq),
                    ack=int(packet.ack),
                )

    def _prune_timeouts(self, *, now_ms: int) -> None:
        timeout = max(250, int(self.cfg.link_timeout_ms))
        for peer_id, peer in list(self._peers_by_id.items()):
            if (int(now_ms) - int(peer.last_seen_ms)) < int(timeout):
                continue
            self.log.info(
                "relay_peer_timeout",
                peer_id=str(peer.peer_id),
                room_code=str(peer.room_code or ""),
                slot_index=int(peer.slot_index),
                timeout_ms=int(timeout),
                silent_ms=int(now_ms) - int(peer.last_seen_ms),
            )
            self._disconnect_peer(peer=peer, reason="timeout", now_ms=int(now_ms))
            self._peers_by_id.pop(str(peer_id), None)
            self._peers_by_addr.pop(peer.addr, None)

    def _disconnect_peer(self, *, peer: _Peer, reason: str, now_ms: int) -> None:
        room = self._rooms.get(str(peer.room_code))
        if room is None:
            self.log.info(
                "relay_peer_disconnected_without_room",
                peer_id=str(peer.peer_id),
                reason=str(reason),
            )
            return
        if 0 <= int(peer.slot_index) < len(room.slots):
            slot = room.slots[int(peer.slot_index)]
            if slot.peer_id == str(peer.peer_id):
                slot.peer_id = ""
                slot.ready = False
                slot.disconnected_ms = int(now_ms)
                slot.peer_name = ""

        self.log.info(
            "relay_peer_disconnected",
            peer_id=str(peer.peer_id),
            room_code=str(room.room_code),
            slot_index=int(peer.slot_index),
            reason=str(reason),
        )
        self._broadcast_room_state(room=room, now_ms=int(now_ms))

        notice = PeerDisconnect(slot_index=int(peer.slot_index), reason=str(reason))
        for slot in room.slots:
            if not slot.connected:
                continue
            dst = self._peers_by_id.get(str(slot.peer_id))
            if dst is None:
                continue
            self._send_peer(dst, notice, reliable=True, now_ms=int(now_ms))

        if all((not slot.connected) for slot in room.slots):
            self._rooms.pop(str(room.room_code), None)
            self.log.info(
                "relay_room_evicted",
                room_code=str(room.room_code),
                session_id=str(room.session_id),
                reason="all_peers_disconnected",
            )

    @staticmethod
    def _next_free_slot(room: _Room) -> int:
        for slot in room.slots:
            if not slot.connected:
                return int(slot.slot_index)
        return -1

    @staticmethod
    def _addr_text(addr: PeerAddr) -> str:
        return f"{addr[0]}:{addr[1]}"


__all__ = ["RelayServer", "RelayServerConfig"]



from __future__ import annotations

import math
from collections.abc import Callable, MutableSequence, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

from grim.geom import Vec2

from ...creatures.damage_types import CreatureDamageType
from ...creatures.spawn import CreatureFlags
from ...math_parity import f32
from ...weapons import weapon_entry_for_projectile_type_id
from ..effects import (
    _spawn_ion_hit_effects,
    _spawn_plasma_cannon_hit_effects,
    _spawn_shrinkifier_hit_effects,
    _spawn_splitter_hit_effects,
)
from ..types import (
    CreatureDamageApplier,
    Damageable,
    Projectile,
    ProjectileRuntimeState,
    ProjectileTypeId,
    _EffectsLike,
)
from .collision import _apply_damage_to_creature, _hit_radius_for

if TYPE_CHECKING:
    from .projectile_pool import ProjectilePool


@dataclass(slots=True)
class _ProjectileUpdateCtx:
    pool: ProjectilePool
    creatures: Sequence[Damageable]
    dt: float
    ion_scale: float
    detail_preset: int
    rng: Callable[[], int]
    runtime_state: ProjectileRuntimeState | None
    effects: _EffectsLike | None
    sfx_queue: MutableSequence[str] | None
    apply_creature_damage: CreatureDamageApplier | None


@dataclass(slots=True)
class _ProjectileHitInfo:
    proj_index: int
    proj: Projectile
    hit_idx: int
    move: Vec2
    target: Vec2


@dataclass(slots=True)
class _ProjectileHitPerkCtx:
    proj: Projectile
    creature: Damageable
    rng: Callable[[], int]
    owner_perk_active: Callable[[int, int], bool]
    poison_idx: int


_ProjectileHitPerkHook = Callable[[_ProjectileHitPerkCtx], None]


def _projectile_hit_perk_poison_bullets(ctx: _ProjectileHitPerkCtx) -> None:
    if ctx.owner_perk_active(int(ctx.proj.owner_id), int(ctx.poison_idx)) and (int(ctx.rng()) & 7) == 1:
        ctx.creature.flags |= CreatureFlags.SELF_DAMAGE_TICK


_PROJECTILE_HIT_PERK_HOOKS: tuple[_ProjectileHitPerkHook, ...] = (_projectile_hit_perk_poison_bullets,)


ProjectileLingerHandler = Callable[[_ProjectileUpdateCtx, Projectile], None]
ProjectilePreHitCreatureHandler = Callable[[_ProjectileUpdateCtx, Projectile, int], None]
ProjectilePostHitCreatureHandler = Callable[[_ProjectileUpdateCtx, _ProjectileHitInfo], None]


@dataclass(frozen=True, slots=True)
class ProjectileBehavior:
    linger: ProjectileLingerHandler
    pre_hit_creature: ProjectilePreHitCreatureHandler | None = None
    post_hit_creature: ProjectilePostHitCreatureHandler | None = None


def _life_timer_sub_f32(life_timer: float, amount: float) -> float:
    return float(f32(float(life_timer) - float(amount)))


def _linger_default(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt))


def _linger_gauss_gun(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt) * 0.1)


def _linger_ion_minigun(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt))
    damage = ctx.dt * 40.0
    radius = ctx.ion_scale * 60.0
    for creature_idx, creature in enumerate(ctx.creatures):
        if not creature.active:
            continue
        if creature.hitbox_size <= 5.0:
            continue
        creature_radius = _hit_radius_for(creature)
        hit_r = radius + creature_radius
        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
            _apply_damage_to_creature(
                ctx.creatures,
                creature_idx,
                damage,
                damage_type=CreatureDamageType.ION,
                impulse=Vec2(),
                owner_id=int(proj.owner_id),
                apply_creature_damage=ctx.apply_creature_damage,
            )


def _linger_ion_rifle(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt))
    damage = ctx.dt * 100.0
    radius = ctx.ion_scale * 88.0
    for creature_idx, creature in enumerate(ctx.creatures):
        if not creature.active:
            continue
        if creature.hitbox_size <= 5.0:
            continue
        creature_radius = _hit_radius_for(creature)
        hit_r = radius + creature_radius
        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
            _apply_damage_to_creature(
                ctx.creatures,
                creature_idx,
                damage,
                damage_type=CreatureDamageType.ION,
                impulse=Vec2(),
                owner_id=int(proj.owner_id),
                apply_creature_damage=ctx.apply_creature_damage,
            )


def _linger_ion_cannon(ctx: _ProjectileUpdateCtx, proj: Projectile) -> None:
    proj.life_timer = _life_timer_sub_f32(float(proj.life_timer), float(ctx.dt) * 0.7)
    damage = ctx.dt * 300.0
    radius = ctx.ion_scale * 128.0
    for creature_idx, creature in enumerate(ctx.creatures):
        if not creature.active:
            continue
        if creature.hitbox_size <= 5.0:
            continue
        creature_radius = _hit_radius_for(creature)
        hit_r = radius + creature_radius
        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
            _apply_damage_to_creature(
                ctx.creatures,
                creature_idx,
                damage,
                damage_type=CreatureDamageType.ION,
                impulse=Vec2(),
                owner_id=int(proj.owner_id),
                apply_creature_damage=ctx.apply_creature_damage,
            )


def _pre_hit_splitter(ctx: _ProjectileUpdateCtx, proj: Projectile, hit_idx: int) -> None:
    _spawn_splitter_hit_effects(
        ctx.effects,
        pos=proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )
    # Native player-hit checks in `projectile_update` key off `owner_id != -100`.
    # Splitter children are spawned with `owner_id = hit creature index`, so they
    # can hit players even when the parent projectile owner was `-100`.
    split_hits_players = int(hit_idx) != -100
    ctx.pool.spawn(
        pos=proj.pos,
        angle=proj.angle - 1.0471976,
        type_id=ProjectileTypeId.SPLITTER_GUN,
        owner_id=int(hit_idx),
        base_damage=proj.base_damage,
        hits_players=split_hits_players,
    )
    ctx.pool.spawn(
        pos=proj.pos,
        angle=proj.angle + 1.0471976,
        type_id=ProjectileTypeId.SPLITTER_GUN,
        owner_id=int(hit_idx),
        base_damage=proj.base_damage,
        hits_players=split_hits_players,
    )


def _post_hit_ion_common(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    _spawn_ion_hit_effects(
        ctx.effects,
        ctx.sfx_queue,
        type_id=int(hit.proj.type_id),
        pos=hit.proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )


def _post_hit_ion_rifle(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    runtime_state = ctx.runtime_state
    creatures = ctx.creatures
    hit_creature = int(hit.hit_idx)
    if (
        runtime_state is not None
        and runtime_state.shock_chain_projectile_id == hit.proj_index
        and 0 <= hit_creature < len(creatures)
    ):
        links_left = int(runtime_state.shock_chain_links_left)
        if links_left > 0 and creatures:
            runtime_state.shock_chain_links_left = links_left - 1

            origin_pos = hit.proj.pos
            min_dist_sq = 100.0 * 100.0

            best_idx = 0
            best_dist_sq = 1e12
            for creature_id, creature in enumerate(creatures):
                if creature_id == hit_creature:
                    continue
                if not creature.active:
                    continue
                d_sq = Vec2.distance_sq(origin_pos, creature.pos)
                if d_sq <= min_dist_sq:
                    continue
                if d_sq < best_dist_sq:
                    best_dist_sq = d_sq
                    best_idx = creature_id

            origin = creatures[hit_creature]
            target = creatures[best_idx]
            angle = (target.pos - origin.pos).to_heading()

            prev_guard = bool(runtime_state.bonus_spawn_guard)
            runtime_state.bonus_spawn_guard = True
            try:
                proj_id = ctx.pool.spawn(
                    pos=origin_pos,
                    angle=angle,
                    type_id=int(hit.proj.type_id),
                    owner_id=hit_creature,
                    base_damage=hit.proj.base_damage,
                )
            finally:
                runtime_state.bonus_spawn_guard = prev_guard
            runtime_state.shock_chain_projectile_id = proj_id
    _post_hit_ion_common(ctx, hit)


def _post_hit_plasma_cannon(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    size = float(creature.size)
    ring_radius = size * 0.5 + 1.0

    plasma_entry = weapon_entry_for_projectile_type_id(int(ProjectileTypeId.PLASMA_RIFLE))
    plasma_meta = (
        float(plasma_entry.projectile_meta)
        if plasma_entry and plasma_entry.projectile_meta is not None
        else hit.proj.base_damage
    )

    runtime_state = ctx.runtime_state
    prev_guard = False
    if runtime_state is not None:
        prev_guard = bool(runtime_state.bonus_spawn_guard)
        runtime_state.bonus_spawn_guard = True
    try:
        for ring_idx in range(12):
            ring_angle = float(ring_idx) * (math.pi / 6.0)
            ring_offset = Vec2.from_angle(ring_angle) * ring_radius
            ctx.pool.spawn(
                pos=hit.proj.pos + ring_offset,
                angle=ring_angle,
                type_id=ProjectileTypeId.PLASMA_RIFLE,
                owner_id=-100,
                base_damage=plasma_meta,
            )
    finally:
        if runtime_state is not None:
            runtime_state.bonus_spawn_guard = prev_guard

    _spawn_plasma_cannon_hit_effects(
        ctx.effects,
        ctx.sfx_queue,
        pos=hit.proj.pos,
        detail_preset=ctx.detail_preset,
    )


def _post_hit_shrinkifier(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    _spawn_shrinkifier_hit_effects(
        ctx.effects,
        pos=hit.proj.pos,
        rng=ctx.rng,
        detail_preset=ctx.detail_preset,
    )

    creature = ctx.creatures[int(hit.hit_idx)]
    new_size = float(creature.size) * 0.65
    creature.size = new_size
    if new_size < 16.0:
        _apply_damage_to_creature(
            ctx.creatures,
            int(hit.hit_idx),
            float(creature.hp) + 1.0,
            damage_type=CreatureDamageType.BULLET,
            impulse=Vec2(),
            owner_id=int(hit.proj.owner_id),
            apply_creature_damage=ctx.apply_creature_damage,
        )
    hit.proj.life_timer = 0.25


def _post_hit_pulse_gun(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    creature.pos = creature.pos + hit.move * 3.0


def _post_hit_plague_spreader(ctx: _ProjectileUpdateCtx, hit: _ProjectileHitInfo) -> None:
    creature = ctx.creatures[int(hit.hit_idx)]
    creature.plague_infected = True


_DEFAULT_BEHAVIOR = ProjectileBehavior(linger=_linger_default)

# Public: used by tests to ensure handler coverage.
PROJECTILE_BEHAVIOR_BY_TYPE_ID: dict[int, ProjectileBehavior] = {
    int(ProjectileTypeId.PISTOL): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.ASSAULT_RIFLE): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.SHOTGUN): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.SUBMACHINE_GUN): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.GAUSS_GUN): ProjectileBehavior(linger=_linger_gauss_gun),
    int(ProjectileTypeId.PLASMA_RIFLE): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.PLASMA_MINIGUN): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.PULSE_GUN): ProjectileBehavior(linger=_linger_default, post_hit_creature=_post_hit_pulse_gun),
    int(ProjectileTypeId.ION_RIFLE): ProjectileBehavior(
        linger=_linger_ion_rifle, post_hit_creature=_post_hit_ion_rifle,
    ),
    int(ProjectileTypeId.ION_MINIGUN): ProjectileBehavior(
        linger=_linger_ion_minigun, post_hit_creature=_post_hit_ion_common,
    ),
    int(ProjectileTypeId.ION_CANNON): ProjectileBehavior(
        linger=_linger_ion_cannon, post_hit_creature=_post_hit_ion_common,
    ),
    int(ProjectileTypeId.SHRINKIFIER): ProjectileBehavior(
        linger=_linger_default, post_hit_creature=_post_hit_shrinkifier,
    ),
    int(ProjectileTypeId.BLADE_GUN): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.SPIDER_PLASMA): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.PLASMA_CANNON): ProjectileBehavior(
        linger=_linger_default, post_hit_creature=_post_hit_plasma_cannon,
    ),
    int(ProjectileTypeId.SPLITTER_GUN): ProjectileBehavior(linger=_linger_default, pre_hit_creature=_pre_hit_splitter),
    int(ProjectileTypeId.PLAGUE_SPREADER): ProjectileBehavior(
        linger=_linger_default, post_hit_creature=_post_hit_plague_spreader,
    ),
    int(ProjectileTypeId.RAINBOW_GUN): _DEFAULT_BEHAVIOR,
    int(ProjectileTypeId.FIRE_BULLETS): _DEFAULT_BEHAVIOR,
}
